"""Tests of wavelet tool.

These tests will download various IDL tplot savefiles from https://github.com/spedas/test_data
They will also create multiple png files that contain IDL and python plots for comparison.
These png files are created inside CONFIG["local_data_dir"].
"""

import os
import unittest
import logging
from numpy.testing import assert_allclose
import numpy as np
from pyspedas import (
    get_data,
    tplot,
    tplot_rename,
    ylim,
    zlim,
    options,
    tplot_options,
    split_vec,
    join_vec,
    store_data,
    del_data,
    time_double,
    wavelet,
    wavelet98,
    wav_data,
    tplot_restore,
    tplot_names,
)
from pyspedas.analysis.tests.wavetest import wavetest
from pyspedas.utilities.config_testing import TESTING_CONFIG, test_data_download_file

# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
analysis_dir = "analysis_tools"
save_dir = os.path.join(output_dir, analysis_dir)
if not os.path.exists(save_dir):
    os.makedirs(save_dir)
# Directory with IDL SPEDAS validation files
validation_dir = TESTING_CONFIG["remote_validation_dir"]


def tc_scales_from_freqs(freqs, omega0=6.0):
    """Convert desired physical frequencies [Hz] to wavelet scales using Torrence & Compo (1998)"""
    fourier_factor = (omega0 + np.sqrt(2 + omega0**2)) / (4 * np.pi)
    return 1.0 / (freqs * fourier_factor)  # No dt here!


def fracyear_to_unix(fyear: float):
    # Convert fractional years yyyy.fff from Torrence and Compo example to Unix timestamps
    iyear = int(fyear)
    ut_start = time_double(str(iyear))
    frac = fyear - iyear
    secs = time_double(str(iyear + 1)) - ut_start
    return ut_start + frac * secs


class TwaveletDataValidation(unittest.TestCase):
    """Compares wavelet results between Python and IDL"""

    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The SPEDAS script that creates the file: general/tools/python_validate/wavelet_python_validate.pro
        """

        # Testing tolerance
        cls.tol = 1e-10

        # Load validation variables from the test file
        filename = test_data_download_file(
            validation_dir, analysis_dir, "wavelet_test.tplot", save_dir
        )
        # If filename does not exist, skip the test
        if not os.path.exists(filename):
            raise unittest.SkipTest(
                "Cannot find local IDL savefile for comparison. Filename: " + filename
            )

        del_data("*")
        tplot_restore(filename)
        tplot_names()

        # Test 1: simple sine waves
        cls.sin_wav = get_data("sin_wav")
        cls.sin_wav_wv_pow = get_data("sin_wav_wv_pow")

        # Test 2: Vassilis homework example using THEMIS FGM data
        cls.tha_fgs_fac_bp_x = get_data("tha_fgs_fac_bp_x")
        cls.tha_fgs_fac_bp_y = get_data("tha_fgs_fac_bp_y")
        cls.tha_fgs_fac_bp_z = get_data("tha_fgs_fac_bp_z")
        cls.tha_fgs_fac_bp_x_wv_pow = get_data("tha_fgs_fac_bp_x_wv_pow")
        cls.tha_fgs_fac_bp_y_wv_pow = get_data("tha_fgs_fac_bp_y_wv_pow")
        cls.tha_fgs_fac_bp_z_wv_pow = get_data("tha_fgs_fac_bp_z_wv_pow")
        cls.wavetest_powspec = get_data("wavetest_powspec")

        tplot_rename("sin_wav_wv_pow", "idl_sin_wav_wv_pow")
        tplot_rename("wavetest_powspec", "idlwavetest_powspec")

        zlim("tha_fgs_fac_bp_?_wv_pow", 1.0e-1, 1.0e2)
        ylim("tha_fgs_fac_bp_?_wv_pow", 1.0e-3, 4.1e-2)
        options("tha_fgs_fac_bp_?_wv_pow", "zlog", True)
        options("tha_fgs_fac_bp_?_wv_pow", "ylog", False)

        tplot_rename("tha_fgs_fac_bp_x_wv_pow", "idl_fgs_fac_bp_x_wv_pow")
        tplot_rename("tha_fgs_fac_bp_y_wv_pow", "idl_fgs_fac_bp_y_wv_pow")
        tplot_rename("tha_fgs_fac_bp_z_wv_pow", "idl_fgs_fac_bp_z_wv_pow")

    def setUp(self):
        """We need to clean tplot variables before each run"""
        # del_data('*')

    def test_sin_wav(self):
        # The default is 'morl' rather than 'cmorl0.5-1.0, but it doesn't appear to make that much difference.
        # cmorl0.5-1.0 is what Eric used in the wiki example.
        wavelet("sin_wav", wavename="cmorl1.5-1.0")
        pvar = "sin_wav_pow"
        options(pvar, "colormap", "jet")
        ylim(pvar, 0.001, 0.1)
        options(pvar, "ylog", True)
        options(pvar, "ytitle", pvar)
        options(pvar, "zlog", True)
        d = get_data("sin_wav_pow")
        logging.info(
            f"Python frequencies: bin count {d.y.shape[1]}, min {np.min(d.v)}, max: {np.max(d.v)}"
        )
        logging.info(f"Python power: min {np.min(d.y)}, max {np.max(d.y)}")
        d_idl = get_data("idl_sin_wav_wv_pow")
        logging.info(
            f"IDL frequencies: bin count {d_idl.y.shape[1]}, min {np.min(d_idl.v)}, max: {np.max(d_idl.v)}"
        )
        logging.info(f"IDL power: min {np.min(d_idl.y)}, max {np.max(d_idl.y)}")

        local_png = os.path.join(save_dir, "sin_wav_simple.png")
        tvar = ["sin_wav", "sin_wav_pow", "idl_sin_wav_wv_pow"]
        tplot(tvar, display=global_display, save_png=local_png)

    def test_equiv_wav(self):

        # 54 logarithmically spaced frequencies between IDL min and max
        idl_freqs = np.logspace(np.log10(0.00051), np.log10(0.05), 54)

        scales = tc_scales_from_freqs(idl_freqs)
        self.assertIsNotNone(scales)

        wavelet("sin_wav", wavename="cmorl1.5-1.0", sampling_period=1.0)
        pvar = "sin_wav_pow"
        options(pvar, "colormap", "spedas")
        ylim(pvar, 0.0005, 0.05)
        options(pvar, "ylog", True)
        options(pvar, "zlog", True)
        options(pvar, "ytitle", pvar)
        d = get_data(pvar)
        logging.info(
            f"Python frequencies: bin count {d.y.shape[1]}, min {np.min(d.v)}, max: {np.max(d.v)}"
        )
        logging.info(f"Python power: min {np.min(d.y)}, max {np.max(d.y)}")
        d_idl = get_data("idl_sin_wav_wv_pow")
        logging.info(
            f"IDL frequencies: bin count {d_idl.y.shape[1]}, min {np.min(d_idl.v)}, max: {np.max(d_idl.v)}"
        )
        logging.info(f"IDL power: min {np.min(d_idl.y)}, max {np.max(d_idl.y)}")

        local_png = os.path.join(save_dir, "sin_wav_equiv.png")
        tvar = ["sin_wav", "sin_wav_pow", "idl_sin_wav_wv_pow"]
        tplot(tvar, display=global_display, save_png=local_png)

    def test_themis_fgm_wavelet(self):
        wavelet("tha_fgs_fac_bp_x")
        wavelet("tha_fgs_fac_bp_y")
        wavelet("tha_fgs_fac_bp_z")

        zlim("tha_fgs_fac_bp_?_pow", 1.0e-1, 1.0e2)
        ylim("tha_fgs_fac_bp_?_pow", 1.0e-3, 4.1e-2)
        options("tha_fgs_fac_bp_?_pow", "zlog", True)
        options("tha_fgs_fac_bp_?_pow", "ylog", False)
        d = get_data("tha_fgs_fac_bp_x_pow")
        logging.info(
            f"Python frequencies: bin count {d.y.shape[1]}, min {np.min(d.v)}, max: {np.max(d.v)}"
        )
        logging.info(f"Python power: min {np.min(d.y)}, max {np.max(d.y)}")
        d_idl = get_data("idl_fgs_fac_bp_x_wv_pow")
        logging.info(
            f"IDL frequencies: bin count {d_idl.y.shape[1]}, min {np.min(d_idl.v)}, max: {np.max(d_idl.v)}"
        )
        logging.info(f"IDL power: min {np.min(d_idl.y)}, max {np.max(d_idl.y)}")

        local_png = os.path.join(save_dir, "themis_fgm_wavelets.png")
        tvar = ["tha_fgs_fac_bp_x_pow", "idl_fgs_fac_bp_x_wv_pow"]
        tplot(tvar, display=global_display, save_png=local_png)

    def test_nino3_tc(self):

        output_dict = wavetest(noplot=True)
        timestamps = [fracyear_to_unix(t) for t in output_dict["time"]]
        store_data(
            "sst_nino3_pwr",
            data={
                "x": timestamps,
                "y": output_dict["power"],
                "v": output_dict["period"],
            },
        )
        options("sst_nino3_pwr", "spec", 1)
        options("sst_nino3_pwr", "yrange", [0.5, 128.0, 1])
        options("sst_nino3_pwr", "zlog", 0)

        options("idlwavetest_powspec", "spec", 1)
        options("idlwavetest_powspec", "yrange", [0.5, 128.0, 1])
        options("idlwavetest_powspec", "zlog", 0)

        local_png = os.path.join(save_dir, "nino_powspec.png")
        tvar = "sst_nino3_pwr idlwavetest_powspec"
        tplot(tvar, display=global_display, save_png=local_png)
        d = get_data("sst_nino3_pwr")
        assert_allclose(d.times, self.wavetest_powspec.times)
        assert_allclose(d.v, self.wavetest_powspec.v)
        # Max relative difference 2.8e-06
        assert_allclose(d.y, self.wavetest_powspec.y, rtol=3e-05)

    def test_wav_data(self):
        # Compare results of wav_data.py to wav_data.pro

        # IDL results (from savefile)
        # IDL Themis data, after all transformation, but before applying wavelets
        tvars = ["tha_fgs_fac_bp_x", "tha_fgs_fac_bp_y", "tha_fgs_fac_bp_z"]
        # IDL results after applying wav_data.pro
        idl_pow = [
            "idl_fgs_fac_bp_x_wv_pow",
            "idl_fgs_fac_bp_y_wv_pow",
            "idl_fgs_fac_bp_z_wv_pow",
        ]

        # Python results (to be computed below)
        py_pow = [
            "tha_fgs_fac_bp_x_wv_pow",
            "tha_fgs_fac_bp_y_wv_pow",
            "tha_fgs_fac_bp_z_wv_pow",
        ]

        all_vars = []
        dcomp = ["time", "y", "v"]
        dxyz = ["x", "y", "z"]

        for i in range(3):
            # Compute the wavelet transform for each of x,y,z
            wav_data(tvars[i])

            # Compare python results with IDL results
            d1 = get_data(idl_pow[i])
            d2 = get_data(py_pow[i])
            logging.info(f"IDL {idl_pow[i]}: {d1.y.shape}, {d1.v.shape}")
            logging.info(f"Python {py_pow[i]}: {d2.y.shape}, {d2.v.shape}")
            for j in range(3):
                # Find max difference
                max_diff = np.abs(np.max(d1[j] - d2[j]))
                logging.info(f"- Max difference for {dcomp[j]}: {max_diff}")
                assert_allclose(d1[j], d2[j], rtol=1e-3, atol=1e-3)

            # Set plotting parameters
            options(py_pow[i], "ytitle", "python " + dxyz[i])
            options(idl_pow[i], "ytitle", "idl " + dxyz[i])
            options(py_pow[i], "ztitle", "python Ptot")
            options(idl_pow[i], "ztitle", "idl Ptot")
            options(py_pow[i], "zlog", True)
            options(py_pow[i], "ylog", False)
            options(idl_pow[i], "zlog", True)
            options(idl_pow[i], "ylog", False)
            zlim(py_pow[i], 1.0e-1, 1.0e2)
            ylim(py_pow[i], 1.0e-3, 4.1e-2)
            zlim(idl_pow[i], 1.0e-1, 1.0e2)
            ylim(idl_pow[i], 1.0e-3, 4.1e-2)
            all_vars.append(idl_pow[i])
            all_vars.append(py_pow[i])

        # Plot 6 panels (x,y,z - IDL, python)
        tplot_options("title", "wav_data: IDL - Python comparison")
        local_png = os.path.join(save_dir, "wav_data_test.png")
        tplot(all_vars, display=global_display, save_png=local_png)

    def test_store_data_singleton_v(self):
        times = [0.0, 1.0, 2.0]
        y = [0.0, 1.0, 2.0]
        v1 = [1]
        v2 = [0.1, 0.2, 0.3]

        # scalar v  (e.g. for 3-vector, or spectrogram with non-time-varying bin values
        store_data("newvar", data={"x": times, "y": y, "v": v1})
        md = get_data("newvar", metadata=True)
        assert_allclose(md["extra_v_values"], np.array(v1))
        # array v (e.g. spectrogram with time-varying bin values)
        store_data("newvar2", data={"x": times, "y": y, "v1": v2})
        md = get_data("newvar2", metadata=True)
        assert_allclose(md["extra_v_values"], np.array(v2))

    def test_join_vec(self):
        dat_x = get_data("tha_fgs_fac_bp_x")
        dat_y = get_data("tha_fgs_fac_bp_y")
        dat_z = get_data("tha_fgs_fac_bp_z")
        self.assertEqual(np.shape(dat_x), np.shape(dat_y))
        self.assertEqual(np.shape(dat_y), np.shape(dat_z))
        join_vec(
            ["tha_fgs_fac_bp_x", "tha_fgs_fac_bp_y", "tha_fgs_fac_bp_z"],
            newname="tha_fgs_joined",
        )
        dat_jv = get_data("tha_fgs_joined")
        self.assertEqual(
            len(dat_jv), 3
        )  # Should have times, yvals, and a 'v' component

    def test_split_rejoin_non_time_varying(self):
        time = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
        yvals = np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 1.0, 1.0],
                [2.0, 2.0, 2.0],
                [3.0, 3.0, 3.0],
                [4.0, 4.0, 4.0],
            ]
        )
        const_v = np.array([11.0, 12.0, 13.0])
        store_data("newvar", data={"x": time, "y": yvals, "v": const_v})
        orig_data = get_data("newvar")
        self.assertIsNotNone(orig_data)
        split_vec("newvar")
        join_vec("newvar_*", newname="newvar_rejoined")
        joined_dat = get_data("newvar_rejoined")
        self.assertEqual(len(joined_dat), 3)
        assert_allclose(joined_dat.v, const_v)

    def test_split_rejoin_time_varying(self):
        time = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
        yvals = np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 1.0, 1.0],
                [2.0, 2.0, 2.0],
                [3.0, 3.0, 3.0],
                [4.0, 4.0, 4.0],
            ]
        )
        time_varying_v = np.array(
            [
                [1.0, 2.0, 3.0],
                [11.0, 12.0, 13.0],
                [21.0, 22.0, 23.0],
                [31.0, 32.0, 33.0],
                [41.0, 42.0, 43.0],
            ]
        )
        store_data("newvar_tv", data={"x": time, "y": yvals, "v": time_varying_v})
        orig_data = get_data("newvar_tv")
        self.assertIsNotNone(orig_data)
        split_vec("newvar_tv")
        dx = get_data("newvar_tv_x")
        self.assertIsNotNone(dx)
        md = get_data("newvar_tv_x", metadata=True)
        self.assertIn("extra_v_values", md)
        join_vec("newvar_tv_*", newname="newvar_tv_rejoined")
        joined_dat = get_data("newvar_tv_rejoined")
        self.assertEqual(len(joined_dat), 3)
        assert_allclose(joined_dat.v, time_varying_v)

    def test_wav_data_keywords(self):
        # Test various wav_data keywords
        # using a synthetic sine wave dataset so that the test is fast and self-contained

        del_data()
        t = np.arange(4000, dtype=float)
        vtime = time_double("2010-01-01") + 10 * t

        # Create a two-dimensional variable
        var = "sin_wav"
        vdata = np.zeros((4000, 2), dtype=float)
        vdata[:, 0] = np.cos(2 * np.pi * t / 32.0)
        vdata[:, 1] = np.sin(2 * np.pi * t / 64.0)
        store_data(var, data={"x": vtime, "y": vdata})

        # Test rbin, dimennum
        wav_data(var, dimennum=1, rbin=4)
        var1x = "sin_wav(1)_wv_pow"
        var1 = "test_rbin"
        tplot_rename(var1x, var1)
        d = get_data(var1)
        # Since rbin=4, length should be 1000
        self.assertEqual(len(d[0]), 1000)

        # Test trange
        wav_data(var, dimennum=0, trange=[vtime[1000], vtime[2000]])
        var2x = "sin_wav(0)_wv_pow"
        var2 = "test_trange"
        tplot_rename(var2x, var2)
        d = get_data(var2)
        # Since trange is 1001 points, length should be 1001
        self.assertEqual(len(d[0]), 1001)

        # Test resolution
        wav_data(var, dimennum=0, resolution=8)
        var3x = "sin_wav(0)_wv_pow"
        var3 = "test_resolution"
        tplot_rename(var3x, var3)
        d = get_data(var3)
        # Since resolution=8, length should be 500
        self.assertEqual(len(d[0]), 500)

        # Test prange
        wav_data(var, dimennum=0, prange=[8, 60])
        var4x = "sin_wav(0)_wv_pow"
        var4 = "test_prange"
        tplot_rename(var4x, var4)
        d = get_data(var4)
        # Period range should be within 8 to 60
        p = np.array(1.0 / d[2])
        logging.info(p)
        self.assertTrue(np.min(p) >= 8.0)
        self.assertTrue(np.max(p) <= 60.0)
        # In this case, the first p should be 8.0
        self.assertAlmostEqual(np.min(p), 8.0, delta=0.01)

        # Plot all results (optional)
        vnames = tplot_names()
        logging.info(vnames)
        local_png = os.path.join(save_dir, "wav_data_keywords.png")
        tplot(vnames, display=global_display, save_png=local_png)

    def test_wav_data_keywords2(self):
        # Test wav_data with various keywords and compare with IDL results.
        # This is using the IDL save file wav_data_key2_test.tplot
        # which was created by the SPEDAS script: wav_data_validate.pro

        logging.info("\n\nTesting wav_data with keywords2\n")

        # Load IDL savefile
        filename = test_data_download_file(
            validation_dir, analysis_dir, "wav_data_key2_test.tplot", save_dir
        )
        # If filename does not exist, skip the test
        if not os.path.exists(filename):
            self.skipTest(
                "Cannot find local IDL savefile for comparison. Filename: " + filename
            )

        # Load validation variables from the test file
        del_data("*")
        tplot_restore(filename)
        tplot_names()

        var = "sin_wav"
        all_vars = [var]
        py_names = [
            "sin_wav_wv_pow",
            "sin_wav_wv_rat_par",
            "sin_wav_wv_pol_par",
            "sin_wav_wv_pol_perp",
        ]

        for n in py_names:
            new_name = "idl_" + n
            tplot_rename(n, new_name)
            all_vars.append(new_name)
            all_vars.append(n)

        wav_data(var, magrat=1, normval=1, fraction=1, kolom=1, rotate_pow=1)

        # Compare with IDL results
        for n in py_names:
            logging.info(f"Comparing {n}")
            d_idl = get_data("idl_" + n)
            d_py = get_data(n)
            ddd = np.abs(d_idl.y - d_py.y)
            logging.info(
                f"Max difference: {np.max(ddd)} at index {np.argmax(ddd)} Mean difference: {np.mean(ddd)}"
            )
            assert_allclose(d_idl.times, d_py.times)
            assert_allclose(d_idl.v, d_py.v)
            # TODO: this fails but probably is OK
            # because the values in the particular example are very small
            # and the images look very similar
            # assert_allclose(d_idl.y, d_py.y, rtol=1e-01, atol=1e-01)

        # Plot all results (optional)
        local_png = os.path.join(save_dir, "wav_data_key2_test.png")
        tplot(all_vars, display=global_display, save_png=local_png)

    def test_wav_data_2d(self):
        # Test wav_data with 2d data and compare with IDL results.
        # This is using the IDL save file wav_data_2d_test.tplot
        # which was created by the SPEDAS script: wav_data_validate.pro

        logging.info("\n\nTesting wav_data 2d\n")

        # Load IDL savefile
        filename = test_data_download_file(
            validation_dir, analysis_dir, "wav_data_2d_test.tplot", save_dir
        )
        # If filename does not exist, skip the test
        if not os.path.exists(filename):
            self.skipTest(
                "Cannot find local IDL savefile for comparison. Filename: " + filename
            )

        # Load validation variables from the test file
        del_data("*")
        tplot_restore(filename)
        tplot_names()

        var = "sin_wav"
        all_vars = [var]
        py_names = [
            "sin_wav_wv_pow",
            "sin_wav_x_wv_pow",
            "sin_wav_y_wv_pow",
            "sin_wav_wv_pol_perp",
        ]
        for n in py_names:
            new_name = "idl_" + n
            tplot_rename(n, new_name)
            all_vars.append(new_name)
            all_vars.append(n)

        wav_data(var, get_components=1)

        # Compare with IDL results
        for n in py_names:
            logging.info(f"Comparing {n}")
            d_idl = get_data("idl_" + n)
            d_py = get_data(n)
            ddd = np.abs(d_idl.y - d_py.y)
            logging.info(
                f"Max difference: {np.max(ddd)} at index {np.argmax(ddd)} Mean difference: {np.mean(ddd)}"
            )
            assert_allclose(d_idl.times, d_py.times)
            assert_allclose(d_idl.v, d_py.v)
            assert_allclose(d_idl.y, d_py.y, rtol=1e-01, atol=1e-01)

        # Plot all results (optional)
        local_png = os.path.join(save_dir, "wav_data_2d_test.png")
        tplot(all_vars, display=global_display, save_png=local_png)

    def test_wav_data_resample(self):
        # Test wav_data resample portion of the code and compare with IDL results.
        # The NaN and varying time steps in the data trigger the resample.
        # This is using the IDL save file wav_data_resample_test.tplot
        # which was created by the SPEDAS script: wav_data_validate.pro

        logging.info("\n\nTesting wav_data resample keyword\n")

        # Load IDL savefile
        filename = test_data_download_file(
            validation_dir, analysis_dir, "wav_data_resample_test.tplot", save_dir
        )
        # If filename does not exist, skip the test
        if not os.path.exists(filename):
            self.skipTest(
                "Cannot find local IDL savefile for comparison. Filename: " + filename
            )

        # Load validation variables from the test file
        del_data("*")
        tplot_restore(filename)
        tplot_names()

        # Rename the IDL results to distinguish them
        tplot_rename("sin_wav_wv_pow", "idl_pow")
        tplot_rename("sin_wav_orig_wv_pow", "idl_orig_pow")

        var = "sin_wav"
        wav_data(var)
        varx = "sin_wav_wv_pow"

        idl_d = get_data("idl_pow")
        py_d = get_data(varx)

        # Compare IDL and Python results
        assert_allclose(idl_d[0], py_d[0], rtol=1e-5, atol=1e-8)
        assert_allclose(idl_d[1], py_d[1], rtol=1e-5, atol=1e-5)
        assert_allclose(idl_d[2], py_d[2], rtol=1e-5, atol=1e-8)

        # Plot all results (optional)
        allvar = [var, varx, "idl_pow"]
        local_png = os.path.join(save_dir, "wav_data_resample_test.png")
        tplot(allvar, display=global_display, save_png=local_png)

    def test_wav_data_cross(self):
        # Test wav_data keywords cross1, cross2 and compare with IDL results.
        # This is using the IDL save file wav_data_cross_test.tplot
        # which was created by the SPEDAS script: wav_data_validate.pro

        logging.info("\n\nTesting wav_data cross1, cross2 keywords\n")

        # Load IDL savefile
        filename = test_data_download_file(
            validation_dir, analysis_dir, "wav_data_cross_test.tplot", save_dir
        )
        # If filename does not exist, skip the test
        if not os.path.exists(filename):
            self.skipTest(
                "Cannot find local IDL savefile for comparison. Filename: " + filename
            )

        # Load validation variables from the test file
        del_data("*")
        tplot_restore(filename)
        tplot_names()

        var = "sin_wav"
        all_vars0 = [var]
        all_vars1 = []
        all_vars2 = []
        py_names0 = [
            "sin_wav_wv_pow",
            "sin_wav_x_wv_pow",
            "sin_wav_y_wv_pow",
            "sin_wav_z_wv_pow",
        ]
        py_names1 = [
            "sin_wav_wv_gam_lin",
            "sin_wav_wv_coin_lin",
            "sin_wav_wv_quad_lin",
            "sin_wav_wv_gam_cir",
            "sin_wav_wv_coin_cir",
            "sin_wav_wv_quad_cir",
        ]
        py_names2 = [
            "sin_wav_wv_gam_pr",
            "sin_wav_wv_coin_pr",
            "sin_wav_wv_quad_pr",
            "sin_wav_wv_gam_pl",
            "sin_wav_wv_coin_pl",
            "sin_wav_wv_quad_pl",
        ]

        # Rename the IDL results to distinguish them
        for n in py_names0:
            new_name = "idl_" + n
            tplot_rename(n, new_name)
            all_vars0.append(new_name)
            all_vars0.append(n)
        for n in py_names1:
            new_name = "idl_" + n
            tplot_rename(n, new_name)
            all_vars1.append(new_name)
            all_vars1.append(n)
        for n in py_names2:
            new_name = "idl_" + n
            tplot_rename(n, new_name)
            all_vars2.append(new_name)
            all_vars2.append(n)

        # Compute the python results
        wav_data(var, get_components=1, cross1=1, cross2=1)

        # Compare base results
        d_idl = get_data("idl_sin_wav_wv_pow")
        d_py = get_data("sin_wav_wv_pow")
        assert_allclose(d_idl.times, d_py.times)
        assert_allclose(d_idl.v, d_py.v)
        assert_allclose(d_idl.y, d_py.y, rtol=1e-03, atol=1e-3)

        # Compare x,y,z results
        for n in py_names0:
            logging.info(f"Comparing {n}")
            d_idl = get_data("idl_" + n)
            d_py = get_data(n)
            ddd = np.abs(d_idl.y - d_py.y)
            logging.info(
                f"Max difference: {np.max(ddd)} at index {np.argmax(ddd)} Mean difference: {np.mean(ddd)}"
            )
            assert_allclose(d_idl.times, d_py.times)
            assert_allclose(d_idl.v, d_py.v)
            assert_allclose(d_idl.y, d_py.y, rtol=7e-03, atol=7e-03)

        # Compare cross1 results
        for n in py_names1:
            logging.info(f"Comparing {n}")
            d_idl = get_data("idl_" + n)
            d_py = get_data(n)
            ddd = np.abs(d_idl.y - d_py.y)
            logging.info(
                f"Max difference: {np.max(ddd)} at index {np.argmax(ddd)} Mean difference: {np.mean(ddd)}"
            )
            assert_allclose(d_idl.times, d_py.times)
            assert_allclose(d_idl.v, d_py.v)
            assert_allclose(d_idl.y, d_py.y, rtol=7e-01, atol=7e-1)

        # Compare cross2 results
        for n in py_names2:
            d_idl = get_data("idl_" + n)
            d_py = get_data(n)
            assert_allclose(d_idl.times, d_py.times)
            assert_allclose(d_idl.v, d_py.v)
            ddd = np.abs(d_idl.y - d_py.y)
            logging.info(
                f"Max difference: {np.max(ddd)} at index {np.argmax(ddd)} Mean difference: {np.mean(ddd)}"
            )
            assert_allclose(d_idl.y, d_py.y, rtol=7e-01, atol=7e-1)

        # Plot all results (optional)
        local_png = os.path.join(save_dir, "wav_data_cross0.png")
        tplot(all_vars0, display=global_display, save_png=local_png)
        local_png = os.path.join(save_dir, "wav_data_cross1.png")
        tplot(all_vars1, display=global_display, save_png=local_png)
        local_png = os.path.join(save_dir, "wav_data_cross2.png")
        tplot(all_vars2, display=global_display, save_png=local_png)

    def test_wavelet98_paul_dog(self):
        # Test wavelet98.py Paul and Dog and compare with IDL results.

        logging.info("\n\nTesting wavelet98 Paul and Dog wavelets\n")

        # Load IDL savefile
        filename = test_data_download_file(
            validation_dir, analysis_dir, "wavelet98_paul_dog_test.tplot", save_dir
        )
        # If filename does not exist, skip the test
        if not os.path.exists(filename):
            self.skipTest(
                "Cannot find local IDL savefile for comparison. Filename: " + filename
            )

        # Load validation variables from the test file
        del_data("*")
        tplot_restore(filename)
        tplot_names()

        # Rename the IDL results to distinguish them
        tplot_rename("wav_paul", "idl_wav_paul")
        options("idl_wav_paul", "ytitle", "idl")
        tplot_rename("pow_paul", "idl_pow_paul")
        options("idl_pow_paul", "ytitle", "idl")
        tplot_rename("wav_dog", "idl_wav_dog")
        options("idl_wav_dog", "ytitle", "idl")
        tplot_rename("pow_dog", "idl_pow_dog")
        options("idl_pow_dog", "ytitle", "idl")

        var = "sin_wav"
        data = get_data(var)
        vtime = data[0]
        vy = data[1]
        dt = np.median(np.diff(vtime))  # time step in seconds

        # PAUL wavelet
        paul_res = wavelet98(vy, dt, mother="PAUL")
        wave_paul = paul_res[0]
        period_paul = paul_res[2]

        store_data("wav_paul", data={"x": vtime, "y": wave_paul, "v": period_paul})
        options("wav_paul", "ytitle", "python")
        options("wav_paul", "ysubtitle", "Paul")

        pow_paul = np.abs(wave_paul) ** 2
        store_data("pow_paul", data={"x": vtime, "y": pow_paul, "v": period_paul})
        options("pow_paul", "ytitle", "Power")
        options("pow_paul", "ysubtitle", "Paul")
        options("pow_paul", "spec", 1)
        options("pow_paul", "ylog", 1)

        # Compare IDL to Python Paul results
        d_idl_paul = get_data("idl_wav_paul")
        d_py_paul = get_data("wav_paul")
        assert_allclose(d_idl_paul.times, d_py_paul.times)
        assert_allclose(d_idl_paul.v, d_py_paul.v)
        assert_allclose(d_idl_paul.y, d_py_paul.y, rtol=7e-05, atol=7e-5)
        p_idl_paul = get_data("idl_pow_paul")
        p_py_paul = get_data("pow_paul")
        assert_allclose(p_idl_paul.times, p_py_paul.times)
        assert_allclose(p_idl_paul.v, p_py_paul.v)
        assert_allclose(p_idl_paul.y, p_py_paul.y, rtol=7e-05, atol=7e-5)

        # DOG wavelet
        dog_res = wavelet98(vy, dt, mother="DOG")
        wave_dog = dog_res[0]
        period_dog = dog_res[2]

        store_data("wav_dog", data={"x": vtime, "y": wave_dog, "v": period_dog})
        options("wav_dog", "ytitle", "python")
        options("wav_dog", "ysubtitle", "Dog")

        pow_dog = np.abs(wave_dog) ** 2
        store_data("pow_dog", data={"x": vtime, "y": pow_dog, "v": period_dog})
        options("pow_dog", "ytitle", "Power")
        options("pow_dog", "ysubtitle", "Dog")
        options("pow_dog", "spec", 1)
        options("pow_dog", "ylog", 1)

        # Compare dog results
        d_idl_dog = get_data("idl_wav_dog")
        d_py_dog = get_data("wav_dog")
        assert_allclose(d_idl_dog.times, d_py_dog.times)
        assert_allclose(d_idl_dog.v, d_py_dog.v)
        assert_allclose(d_idl_dog.y, d_py_dog.y, rtol=7e-05, atol=7e-5)
        p_idl_dog = get_data("idl_pow_dog")
        p_py_dog = get_data("pow_dog")
        assert_allclose(p_idl_dog.times, p_py_dog.times)
        assert_allclose(p_idl_dog.v, p_py_dog.v)
        assert_allclose(p_idl_dog.y, p_py_dog.y, rtol=7e-05, atol=7e-5)

        # Plot all results (optional)
        allvar = [
            var,
            "wav_paul",
            "idl_wav_paul",
            "pow_paul",
            "idl_pow_paul",
            "wav_dog",
            "idl_wav_dog",
            "pow_dog",
            "idl_pow_dog",
        ]
        local_png = os.path.join(save_dir, "wavelet98_paul_dog_test.png")
        tplot(allvar, display=global_display, save_png=local_png)


if __name__ == "__main__":
    unittest.main()



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG

def load(trange=['2020-06-01', '2020-06-02'], 
         instrument='mag',
         datatype='rtn-normal', 
         mode=None,
         level='l2', 
         suffix='', 
         prefix='',
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False,
         force_download=False):
    """
    This function loads data from the Solar Orbiter mission; this function is not meant 
    to be called directly; instead, see the wrappers:
        pyspedas.projects.solo.mag
        pyspedas.projects.solo.epd
        pyspedas.projects.solo.rpw
        pyspedas.projects.solo.swa

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2020-06-01', '2020-06-02']

        instrument: str
            Spacecraft identifier ('mag', 'epd', 'rpw', 'swa')
            Default: 'mag'

        datatype: str
            Valid options: 'rtn-normal'
            Default: 'rtn_normal'

        mode: str
            Valid options: None
            Default: None

        level: str
            Valid options: 'l2'
            Default: 'l2'

        suffix: str
            The tplot variable names will be given this suffix.
            no prefix is added.
            Default: ''
        
        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
            Default: ''

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: 'False', only loads in data with a "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None, all variables are loaded in.

        varnames: list of str
            List of variable names to load
            Default: [], all data variables are loaded

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        force_download: bool
            If True, downloads the file even if a newer version exists locally. 
            Default: False.

    Returns
    ----------
        List of tplot variables created.

    Example
    ----------
        import pyspedas
        from pyspedas.tplot_tools import tplot
        mag_solar_vars = pyspedas.solar.mag(trange=['2020-06-01', '2020-06-02'])

        epd_solar_vars = pyspedas.solar.epd(trange=['2020-06-01', '2020-06-02'])

        rpw_solar_vars = pyspedas.solar.rpw(trange=['2020-06-01', '2020-06-02'])

        swa_solar_vars = pyspedas.solar.swa(trange=['2020-06-01', '2020-06-02'])

    """

    if prefix is None:
        prefix = ''
    
    if suffix is None:
        suffix = ''

    # Defaults for L2, L3 data
    science_or_low_latency = 'science'
    date_format = '%Y%m%d'
    cdf_version = '??'

    res = 24*3600.

    if level == 'll02':
        science_or_low_latency = 'low_latency'
        date_format = '%Y%m%dt??????-*'
        cdf_version = '???'
        res = 86400.0

    if instrument == 'mag':
        if level == 'll02':
            pathformat = instrument+'/'+science_or_low_latency+'/'+level+'/%Y/solo_'+level+'_'+instrument+'_'+date_format+'_v'+cdf_version+'.cdf'
        else:
            pathformat = instrument+'/'+science_or_low_latency+'/'+level+'/'+datatype+'/%Y/solo_'+level+'_'+instrument+'-'+datatype+'_'+date_format+'_v'+cdf_version+'.cdf'
    elif instrument == 'epd':
        pathformat = instrument+'/'+science_or_low_latency+'/'+level+'/'+datatype+'/'+mode+'/%Y/solo_'+level+'_'+instrument+'-'+datatype+'-'+mode+'_'+date_format+'_v'+cdf_version+'.cdf'
    elif instrument == 'rpw':
        pathformat = instrument+'/'+science_or_low_latency+'/'+level+'/'+datatype+'/%Y/solo_'+level+'_'+instrument+'-'+datatype+'_'+date_format+'_v'+cdf_version+'.cdf'
    elif instrument == 'swa':
        if level == 'l2' or level == 'll02':
            if datatype == 'pas-eflux' or datatype == 'pas-grnd-mom' or datatype == 'pas-vdf':
                pathformat = instrument+'/'+science_or_low_latency+'/'+level+'/'+datatype+'/%Y/solo_'+level+'_'+instrument+'-'+datatype+'_'+date_format+'_v'+cdf_version+'.cdf'
            else:
                date_format = '%Y%m%dt??????-*'
                res = 86400.0
                pathformat = instrument+'/'+science_or_low_latency+'/'+level+'/'+datatype+'/%Y/solo_'+level+'_'+instrument+'-'+datatype+'_'+date_format+'_v'+cdf_version+'.cdf'
        elif level == 'l1':
            if datatype == 'his-pha' or datatype == 'his-sensorrates' or datatype == 'pas-3d' or datatype == 'pas-cal' or datatype == 'pas-mom':
                pathformat = instrument+'/'+science_or_low_latency+'/'+level+'/'+datatype+'/%Y/solo_'+level+'_'+instrument+'-'+datatype+'_'+date_format+'_v'+cdf_version+'.cdf'
            else:
                date_format = '%Y%m%dt??????-*'
                res = 86400.0
                pathformat = instrument+'/'+science_or_low_latency+'/'+level+'/'+datatype+'/%Y/solo_'+level+'_'+instrument+'-'+datatype+'_'+date_format+'_v'+cdf_version+'.cdf'
        elif level == 'l3':
            pathformat = instrument+'/'+science_or_low_latency+'/'+level+'/'+datatype+'/%Y/solo_'+level+'_'+instrument+'-'+datatype+'_'+date_format+'_v'+cdf_version+'.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange, res=res)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update, force_download=force_download)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, prefix=prefix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import os
import unittest
from pyspedas.tplot_tools import data_exists, del_data
import pyspedas


class Fast_Tests(unittest.TestCase):
    def test_downloadonly(self):
        files = pyspedas.projects.fast.acf(
            trange=["1999-09-22", "1999-09-23"],
            time_clip=True,
            level="k0",
            downloadonly=True,
        )
        self.assertTrue(os.path.exists(files[0]))

    def test_load_dcf_data(self):
        dcf_vars = pyspedas.projects.fast.dcf(trange=["1999-09-22", "1999-09-23"], level="k0")
        self.assertTrue(data_exists("fast_dcf_BX"))
        self.assertTrue(data_exists("fast_dcf_BY"))
        self.assertTrue(data_exists("fast_dcf_BZ"))
        self.assertTrue("fast_dcf_BZ" in dcf_vars)

    def test_load_dcf_data_l2(self):
        dcf_vars = pyspedas.projects.fast.dcf(trange=["1998-09-22", "1998-09-23"], level="l2")
        self.assertTrue(data_exists("fast_dcf_DeltaB_GEI"))
        self.assertTrue("fast_dcf_DeltaB_GEI" in dcf_vars)

    def test_load_acf_data(self):
        dcf_vars = pyspedas.projects.fast.acf(
            trange=["1999-09-22", "1999-09-23"], time_clip=True, level="k0"
        )
        self.assertTrue(data_exists("fast_acf_HF_PWR"))
        self.assertTrue(data_exists("fast_acf_HF_E_SPEC"))
        self.assertTrue("fast_acf_HF_E_SPEC" in dcf_vars)

    def test_load_esa_data(self):
        esa_vars = pyspedas.projects.fast.esa(
            notplot=True, trange=["1998-09-05/02:00", "1998-09-05/02:30"]
        )
        self.assertTrue("fast_esa_eflux" in esa_vars)

    def test_load_teams_data(self):
        teams_vars = pyspedas.projects.fast.teams(
            trange=["1998-09-05", "1998-09-06"], level="k0"
        )
        self.assertTrue(data_exists("fast_teams_H+"))
        self.assertTrue(data_exists("fast_teams_O+"))
        self.assertTrue(data_exists("fast_teams_He+"))
        self.assertTrue("fast_teams_fa_spin_dec" in teams_vars)

    def test_load_teams_data_l2(self):
        teams_vars = pyspedas.projects.fast.teams(
            trange=["2004-06-01", "2004-06-02"], level="l2"
        )
        self.assertTrue(data_exists("fast_teams_proton_distribution"))
        self.assertTrue("fast_teams_helium_distribution" in teams_vars)

    def test_load_invalid_trange(self):
        del_data()
        t_all = pyspedas.projects.fast.load(trange="bad input")
        self.assertTrue(len(t_all) == 0)
        del_data()
        t_all = pyspedas.projects.fast.load(trange=["2018-12-15"])
        self.assertTrue(len(t_all) == 0)
        del_data()
        t_all = pyspedas.projects.fast.load(trange=["2018-12-15", "2018-12-14"])
        self.assertTrue(len(t_all) == 0)

    def test_load_invalid_instrument(self):
        t_all = pyspedas.projects.fast.load(instrument="bad input")
        self.assertTrue(len(t_all) == 0)


if __name__ == "__main__":
    unittest.main()



import unittest

import pyspedas

import pyspedas
from pyspedas.projects.themis.analysis.scpot2dens import dens_pot
from pyspedas.projects.themis.analysis import scpot2dens
from pyspedas import get_data, tplot_copy
import numpy as np
from numpy.testing import assert_allclose
global_display=False


class TestScpo2densValidation(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The IDL script that creates data file: general/tools/python_validate/thm_state_validate.pro
        """
        from pyspedas.utilities.download import download
        from pyspedas.projects.themis.config import CONFIG
        from pyspedas import tplot_restore, del_data,options

        # Testing time range
        cls.t = ['2008-03-23', '2008-03-28']


        # Download validation file
        remote_server = 'https://github.com/spedas/test_data/raw/refs/heads/main/'
        remote_name = 'themis/thm_scpot_validate.tplot'
        datafile = download(remote_file=remote_name,
                           remote_path=remote_server,
                           local_path=CONFIG['local_data_dir'],
                           no_download=False)
        if not datafile:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        # Load validation variables from the test file
        del_data('*')
        filename = datafile[0]
        tplot_restore(filename)
        for probe in ['tha', 'thb', 'thc','thd','the']:
            tplot_copy(probe+'_scpot2dens',new_name=probe+'_idl_scpot2dens')
            tplot_copy(probe+'_peer_density_npot',new_name=probe+'_idl_scpot2dens_n')
            options(probe+'_idl_scpot2dens','labels',probe+' IDL scpot2dens')
            options(probe+'_idl_scpot2dens','ylog',1)
            options(probe+'_idl_scpot2dens_n', 'labels',probe+' IDL nishi scpot2dens')


    def test_dens_pot_values(self):
        """ Test supporting function dens_pot for initial values"""
        d1 = dens_pot(-1, 0)
        d2 = dens_pot(-1, 1)

        idl_d1 = 3048.9753
        idl_d2 = 34381.586

        self.assertAlmostEqual(d1, idl_d1, delta=0.1)
        self.assertAlmostEqual(d2, idl_d2, delta=0.1)

    def test_dens_pot_pder_values(self):
        """ Test supporting function dens_pot for derivative values"""
        _, d1 = dens_pot(-1, 0, True)
        _, d2 = dens_pot(-1, 1, True)

        idl_d1 = 6278.0029
        idl_d2 = 95702.727

        self.assertAlmostEqual(d1, idl_d1, delta=0.1)
        self.assertAlmostEqual(d2, idl_d2, delta=0.1)

    def test_dens_pot_list_values(self):
        """ Test supporting function dens_pot for tuple of values and derivative"""
        r1, d1 = dens_pot(np.array([0, -1]), 1, True)

        idl_r1 = np.array([3048.9753, 34381.586])
        idl_d1 = np.array([6278.0029, 95702.727])

        for p, idl in zip(r1, idl_r1):
            self.assertAlmostEqual(p, idl, delta=0.1)

        for p, idl in zip(d1, idl_d1):
            self.assertAlmostEqual(p, idl, delta=0.1)

    def test_scpot2dens_c(self):
        """
        Test satellite case of retrieving density with the change of the spacecraft potential offset.
        The offset if setup for probe c on (2007-7-20/17:24)
        """

        trange = ['2007-7-20/17:23:55', '2007-7-20/17:24:05']

        idl_Npot = np.array([19.2729, 53.6906, 53.0732])

        self.helper_scpot2dens('c', trange, idl_Npot)

    def test_scpot2dens_a(self):
        """
        Test satellite case of retrieving density using data from another spacecraft
        """

        trange = ['2012-01-01/18:10:05', '2012-01-01/18:10:10']

        idl_Npot = np.array([8.74609, 11.7515])

        self.helper_scpot2dens('a', trange, idl_Npot)

    def helper_scpot2dens(self, probe, trange, idl_Npot):
        """
        Helper function that evaluates scpot2dens for diffrent probes and tranges

        Parameters
        ----------
        probe: str
            THEMIS probe
        trange: list[str]
            trange string of the time tange
        idl_Npot: ndarray
            IDL values for the exact same call

        Returns
        -------
        None

        Notes
        -----
        The code that should be executed in IDL to obtain the  idl_Npot values is this:
        ```IDL
        trange = ['2012-01-01/18:10:05']
        probe = 'a'
        thm_load_esa,trange=trange, probe=probe, datat=' peer_avgtemp pe?r_density peer_sc_pot ', level=2
        ; Extract data:
        get_data,'th'+probe+'_peer_density',data=d
        dens_e= d.y
        dens_e_time= d.x
        ;
        get_data,'th'+probe+'_peir_density',data=d
        dens_i= d.y
        dens_i_time= d.x
        ;
        get_data,'th'+probe+'_peer_sc_pot',data=d
        sc_pot = d.y
        sc_pot_time = d.x
        ;
        get_data,'th'+probe+'_peer_avgtemp',data=d
        Te = d.y
        Te_time = d.x

        Npot = thm_scpot2dens(sc_pot, sc_pot_time, Te, Te_time, dens_e, dens_e_time, dens_i, dens_i_time, probe)
        print, Npot
        ```
        """

        pyspedas.projects.themis.esa(trange=trange, probe=probe,
                            varnames=[f'th{probe}_peer_density',
                                      f'th{probe}_peir_density',
                                      f'th{probe}_peer_sc_pot',
                                      f'th{probe}_peer_avgtemp'],
                            level='l2', time_clip=True)

        dens_e_time, dens_e = get_data(f'th{probe}_peer_density')
        dens_i_time, dens_i = get_data(f'th{probe}_peir_density')
        sc_pot_time, sc_pot = get_data(f'th{probe}_peer_sc_pot')
        Te_time, Te = get_data(f'th{probe}_peer_avgtemp')

        Npot = scpot2dens(sc_pot, sc_pot_time, Te, Te_time, dens_e, dens_e_time, dens_i, dens_i_time, probe=probe)

        for p, idl in zip(Npot, idl_Npot):
            self.assertAlmostEqual(p, idl, delta=0.1)

    def test_scpot2dens_idl_a(self):

        trange = ['2014-09-22', '2014-09-23']
        # Set a trange where data is well-behaved enough to make a good test
        nice_trange = ['2014-09-22/08:00', '2014-09-22/19:00']

        probe = 'a'
        self.calc_scpot2dens_single_probe(nice_trange, probe, trange)

        idl_std=get_data('th' + probe + '_idl_scpot2dens-tclip')
        py_std= get_data('th' + probe + '_Npot-tclip')
        idl_nishi = get_data('th' + probe + '_idl_scpot2dens_n-tclip')
        py_nishi = get_data('th' + probe + '_Npot_N-tclip')
        assert_allclose(idl_std.y, py_std.y)
        assert_allclose(idl_nishi.y, py_nishi.y)

    def test_scpot2dens_idl_b(self):

        trange = ['2014-09-22', '2014-09-23']
        # Set a trange where data is well-behaved enough to make a good test
        nice_trange = ['2014-09-22/08:00', '2014-09-22/19:00']
        probe = 'b'
        get_data = self.calc_scpot2dens_single_probe(nice_trange, probe, trange)

        idl_std=get_data('th' + probe + '_idl_scpot2dens-tclip')
        py_std= get_data('th' + probe + '_Npot-tclip')
        idl_nishi = get_data('th' + probe + '_idl_scpot2dens_n-tclip')
        py_nishi = get_data('th' + probe + '_Npot_N-tclip')
        assert_allclose(idl_std.y, py_std.y)
        assert_allclose(idl_nishi.y, py_nishi.y)

    def test_scpot2dens_idl_c(self):

        trange = ['2014-09-22', '2014-09-23']
        # Set a trange where data is well-behaved enough to make a good test
        nice_trange = ['2014-09-22/08:00', '2014-09-22/19:00']

        probe = 'c'
        get_data = self.calc_scpot2dens_single_probe(nice_trange, probe, trange)

        idl_std=get_data('th' + probe + '_idl_scpot2dens-tclip')
        py_std= get_data('th' + probe + '_Npot-tclip')
        idl_nishi = get_data('th' + probe + '_idl_scpot2dens_n-tclip')
        py_nishi = get_data('th' + probe + '_Npot_N-tclip')
        assert_allclose(idl_std.y, py_std.y)
        assert_allclose(idl_nishi.y, py_nishi.y)

    def test_scpot2dens_idl_d(self):

        trange = ['2014-09-22', '2014-09-23']
        # Set a trange where data is well-behaved enough to make a good test
        nice_trange = ['2014-09-22/13:00', '2014-09-22/23:00']

        probe = 'd'
        get_data = self.calc_scpot2dens_single_probe(nice_trange, probe, trange)

        idl_std=get_data('th' + probe + '_idl_scpot2dens-tclip')
        py_std= get_data('th' + probe + '_Npot-tclip')
        idl_nishi = get_data('th' + probe + '_idl_scpot2dens_n-tclip')
        py_nishi = get_data('th' + probe + '_Npot_N-tclip')
        assert_allclose(idl_std.y, py_std.y)
        assert_allclose(idl_nishi.y, py_nishi.y)

    def test_scpot2dens_idl_e(self):

        trange = ['2014-09-22', '2014-09-23']
        # Set a trange where data is well-behaved enough to make a good test
        nice_trange = ['2014-09-22/11:00', '2014-09-22/23:00']

        probe = 'e'
        get_data = self.calc_scpot2dens_single_probe(nice_trange, probe, trange)

        idl_std=get_data('th' + probe + '_idl_scpot2dens-tclip')
        py_std= get_data('th' + probe + '_Npot-tclip')
        idl_nishi = get_data('th' + probe + '_idl_scpot2dens_n-tclip')
        py_nishi = get_data('th' + probe + '_Npot_N-tclip')
        assert_allclose(idl_std.y, py_std.y)
        assert_allclose(idl_nishi.y, py_nishi.y)

    def calc_scpot2dens_single_probe(self, nice_trange, probe, trange):
        from pyspedas.projects.themis.analysis import scpot2dens_nishimura
        from pyspedas.projects.themis.analysis import scpot2dens
        from pyspedas.tplot_tools import get_data
        from pyspedas.tplot_tools import store_data
        from pyspedas.tplot_tools import options
        from pyspedas.tplot_tools import tplot
        import numpy as np

        # get ESA data
        pyspedas.projects.themis.esa(trange=trange, probe=probe,
                                     varnames=['th' + probe + '_peir_density',
                                               'th' + probe + '_peer_sc_pot',
                                               'th' + probe + '_peer_vthermal',
                                               'th' + probe + '_peer_density',
                                               'th' + probe + '_peer_avgtemp',
                                               'th' + probe + '_peer_data_quality'], level='l2')
        # State data
        pyspedas.projects.themis.state(probe=probe, trange=trange)
        pos_data = get_data('th' + probe + '_pos_gsm')
        pos_gsm_time = pos_data.times
        pos_gsm = pos_data.y

        # FGM data
        pyspedas.projects.themis.fgm(trange=trange, probe=probe,
                                     varnames=['th' + probe + '_fgs_gsm'], level='l2')

        # Densities
        dens_i_time, dens_i = get_data('th' + probe + '_peir_density')
        dens_e_time, dens_e = get_data('th' + probe + '_peer_density')
        # SC Potential
        sc_pot_time, sc_pot = get_data('th' + probe + '_peer_sc_pot')
        # Electron Thermal Velocity
        vth_time, vth = get_data('th' + probe + '_peer_vthermal')
        # Average Electron Temperature
        Te_time, Te = get_data('th' + probe + '_peer_avgtemp')
        # Npot Nishimura, no_interp should usually be true if all of the data have the same mode (All ESA data used here are reduced mode).
        Npot_N = scpot2dens_nishimura(sc_pot, sc_pot_time, vth, vth_time, dens_i, dens_i_time, pos_gsm, pos_gsm_time,
                                      probe=probe, no_interp=True)
        # Npot scpot2dens
        Npot = scpot2dens(sc_pot, sc_pot_time, Te, Te_time, dens_e, dens_e_time, dens_i, dens_i_time, probe)

        store_data('th' + probe + '_Npot', {"x": sc_pot_time, "y": Npot})
        store_data('th' + probe + '_Npot_N', {"x": sc_pot_time, "y": Npot_N})
        Npot = 'th' + probe + '_Npot'
        Npot_N = 'th' + probe + '_Npot_N'
        options('th' + probe + '_Npot', 'ylog', 1)
        options('th' + probe + '_Npot_N', 'ylog', 1)
        options('th' + probe + '_Npot', 'yrange', [0.001, 10000.0])
        options('th' + probe + '_Npot_N', 'yrange', [0.001, 10000.0])
        options('th' + probe + '_peer_density', 'yrange', [0.001, 10000.0])
        options('th' + probe + '_peir_density', 'yrange', [0.001, 10000.0])
        options('th' + probe + '_Npot', 'color', 'green')
        options('th' + probe + '_Npot_N', 'color', 'red')
        options('th' + probe + '_peer_density', 'color', 'blue')
        options('th' + probe + '_peir_density', 'color', 'orange')
        options('th' + probe + '_Npot', 'labels', 'N_{scpot}')
        options('th' + probe + '_Npot_N', 'labels', 'N_{scpot}_N')
        options('th' + probe + '_peer_density', 'labels', 'N_e')
        options('th' + probe + '_peir_density', 'labels', 'N_i')

        # Make a multi-tplot variable
        store_data('th' + probe + '_N_comp', ['th' + probe + '_Npot',
                                              'th' + probe + '_Npot_N',
                                              'th' + probe + '_peer_density',
                                              'th' + probe + '_peir_density'])

        # other plots, first L value
        R_E = 6371.0
        pos_l = np.linalg.norm(pos_gsm / R_E, 2, 1)
        store_data('th' + probe + '_Lval', {"x": pos_gsm_time, "y": pos_l})
        options('th' + probe + '_Lval', 'labels', 'L')
        # scpot
        options('th' + probe + '_peer_sc_pot', 'labels', 'SC_POT')
        # Electron Temperature
        options('th' + probe + '_peer_avgtemp', 'labels', 'Elec_Temp')
        options('th' + probe + '_peer_avgtemp', 'ylog', 0)
        # Quality flags
        options('th' + probe + '_peer_data_quality', 'labels', 'DQ')
        # FGM
        options('th' + probe + '_fgs_gsm', 'labels', ['FGS_x', 'FGS_y', 'FGS_z'])
        options('th' + probe + '_fgs_gsm', 'yrange', [-100.0, 100.0])

        if isinstance(trange[0], str):
            fname = 'scpot2dens_test_' + probe + '_' + trange[0]
        else:
            fname = 'scpot2dens_test_' + probe + '_' + pyspedas.time_string(trange[0])

        npdat = get_data('th' + probe + '_Npot')
        np_n_dat = get_data('th' + probe + '_Npot_N')
        idl_npdat = get_data('th' + probe + '_idl_scpot2dens')
        idl_np_n_dat = get_data('th' + probe + '_idl_scpot2dens_n')
        store_data('th' + probe + '_comp_std', ['th' + probe + '_Npot', 'th' + probe + '_idl_scpot2dens'])
        options('th' + probe + '_comp_std', 'color', ['b', 'r'])
        options('th' + probe + '_comp_std', 'ylog', 1)
        store_data('th' + probe + '_comp_nishi', ['th' + probe + '_Npot_N', 'th' + probe + '_idl_scpot2dens_n'])
        options('th' + probe + '_comp_nishi', 'color', ['b', 'r'])

        tplot(['th' + probe + '_Lval',
               'th' + probe + '_peer_sc_pot',
               'th' + probe + '_peer_avgtemp',
               'th' + probe + '_N_comp',
               'th' + probe + '_peer_data_quality',
               'th' + probe + '_fgs_gsm',
               'th' + probe + '_idl_scpot2dens',
               'th' + probe + '_idl_scpot2dens_n',
               'th' + probe + '_comp_std',
               'th' + probe + '_comp_nishi'],
              save_png=fname, display=global_display)

        # Time clip outputs to nice range
        pyspedas.time_clip(
            ['th' + probe + '_idl_scpot2dens', 'th' + probe + '_idl_scpot2dens_n', 'th' + probe + '_Npot',
             'th' + probe + '_Npot_N'], nice_trange[0], nice_trange[1])
        return get_data


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pyspedas
import fnmatch
import logging


def del_data(name=None):
    """
    This function will delete tplot variables that are already stored in memory.  
    
    Parameters
    ----------
        name : str or list[str]
            Names of the tplot variables to be deleted.  If no name is provided, then
            all tplot variables will be deleted.  (wildcards accepted)
         
    Returns
    -------
        None
        
    Examples
    --------
        >>> # Delete Variable 1
        >>> import pyspedas
        >>> pyspedas.del_data("Variable1")

    """
    if name is None:
        name = '*'

    names=pyspedas.tnames(name)
    if len(names) < 1:
        logging.warning("del_data: No valid tplot variables found, returning")
        return

    for name in names:
        if isinstance(pyspedas.tplot_tools.data_quants[name], dict):
            temp_data_quants = pyspedas.tplot_tools.data_quants[name]
            str_name = temp_data_quants['name']
        else:
            temp_data_quants = pyspedas.tplot_tools.data_quants[name]
            str_name = temp_data_quants.name

        del pyspedas.tplot_tools.data_quants[str_name]
        
    return


import os
import unittest
from pyspedas.tplot_tools import data_exists, del_data
import pyspedas

class LoadTestCases(unittest.TestCase):
    def test_utc_timestamp_regression(self):
        varname = 'BX_GSE'
        del_data('*')
        data_omni = pyspedas.projects.omni.data(trange=['2010-01-01/00:00:00', '2010-01-02/00:00:00'],notplot=True,varformat=varname,time_clip=True)
        ts=data_omni[varname]['x'][0]
        # Depending on the version of cdflib in use, ts might be a datetime.datetime, or np.datetime64.  The string representations
        # are slightly different: "2010-01-01 00:00:00" vs "2010-01-01T00:00:00.000000".  So we replace "T" with " " and
        # discard any fractional part before comparing.
        ds=str(ts)
        ds2=ds.replace("T"," ")
        nofrac,frac=ds2.split(".")
        self.assertTrue(nofrac == '2010-01-01 00:00:00')

    def test_load_hro2_data(self):
        del_data('*')
        omni_vars = pyspedas.projects.omni.data()
        self.assertTrue(data_exists('BX_GSE'))
        self.assertTrue(data_exists('BY_GSE'))
        self.assertTrue(data_exists('BZ_GSE'))
        self.assertTrue(data_exists('BY_GSM'))
        self.assertTrue(data_exists('BZ_GSM'))
        self.assertTrue(data_exists('proton_density'))

    def test_load_hro_data(self):
        del_data('*')
        omni_vars = pyspedas.projects.omni.data(level='hro')
        self.assertTrue(data_exists('BX_GSE'))
        self.assertTrue(data_exists('BY_GSE'))
        self.assertTrue(data_exists('BZ_GSE'))
        self.assertTrue(data_exists('BY_GSM'))
        self.assertTrue(data_exists('BZ_GSM'))
        self.assertTrue(data_exists('proton_density'))

    def test_load_hro_5min_data(self):
        del_data('*')
        omni_vars = pyspedas.projects.omni.data(level='hro', datatype='5min', prefix='omni_', suffix='_test')
        self.assertTrue(data_exists('omni_BX_GSE_test'))
        self.assertTrue(data_exists('omni_BY_GSE_test'))
        self.assertTrue(data_exists('omni_BZ_GSE_test'))
        self.assertTrue(data_exists('omni_BY_GSM_test'))
        self.assertTrue(data_exists('omni_BZ_GSM_test'))
        self.assertTrue(data_exists('omni_proton_density_test'))

    def test_load_hro_hour_data(self):
        del_data('*')
        omni_vars = pyspedas.projects.omni.data(level='hro2', datatype='hour', trange=['2013-03-01', '2013-03-02'])
        self.assertTrue(data_exists('BX_GSE'))
        self.assertTrue(data_exists('BY_GSE'))
        self.assertTrue(data_exists('BZ_GSE'))
        self.assertTrue(data_exists('BY_GSM'))
        self.assertTrue(data_exists('BZ_GSM'))

    def test_load_hro_hour_data_long_interval(self):
        from pyspedas.utilities.dailynames import dailynames
        del_data('*')
        trange=['2013-01-05','2013-11-06']
        pathformat = 'hourly/%Y/omni2_h0_mrg1hr_%Y%m01_v??.cdf'
        file_res=24*3600*183 # 1 file every 6 months
        remote_names=dailynames(file_format=pathformat,trange=trange,res=file_res)
        print(remote_names)
        omni_vars = pyspedas.projects.omni.data(datatype='hour', trange=trange)
        self.assertTrue(data_exists('BX_GSE'))
        self.assertTrue(data_exists('BY_GSE'))
        self.assertTrue(data_exists('BZ_GSE'))
        self.assertTrue(data_exists('BY_GSM'))
        self.assertTrue(data_exists('BZ_GSM'))

    def test_load_invalid_datatype(self):
        omni_vars = pyspedas.projects.omni.data(datatype='1')

    def test_downloadonly(self):
        files = pyspedas.projects.omni.data(downloadonly=True, trange=['2014-2-15', '2014-2-16'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import numpy as np
from pyspedas.tplot_tools import data_exists
import pyspedas

global_display = False

class LoadTestCases(unittest.TestCase):
    def test_unpublished_data(self):
        """
            this test doesn't load any data, since the username/pw is invalid
        """
        # no password
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_RTN', username='hello')
        # invalid password
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_RTN', username='hello', password='world')
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_SC', username='hello', password='world')
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_SC_1min', username='hello', password='world')
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_RTN_1min', username='hello', password='world')
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_RTN_4_Sa_per_Cyc', username='hello', password='world')
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_SC_4_Sa_per_Cyc', username='hello', password='world')
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='sqtn_rfs_V1V2', username='hello', password='world')
        spc = pyspedas.projects.psp.spc(trange=['2018-11-5', '2018-11-5/06:00'], username='hello', password='world')
        spi = pyspedas.projects.psp.spi(trange=['2018-11-5', '2018-11-5/06:00'], username='hello', password='world')

    def test_fields_berkeley_multiple_versions(self):
        trange = ['2025-01-01/12:00:00', '2025-01-01/13:00:00']

        fields_vars = pyspedas.projects.psp.fields(trange=trange,
                   datatype='mag_RTN_4_Sa_per_Cyc',
                   time_clip=True,
                   username='none',
                   password='none')

        var_dat = pyspedas.get_data('psp_fld_l2_mag_RTN_4_Sa_per_Cyc')

        # If multiple versions are loaded in the same call, times will be non-monotonic
        t = var_dat.times
        dt = t[1:]-t[0:-1]
        dtmin = np.min(dt)
        self.assertTrue(dtmin > 0.0)

    def test_fields_berkeley_explicit_version(self):
        trange = ['2025-01-01/12:00:00', '2025-01-01/13:00:00']

        fields_vars = pyspedas.projects.psp.fields(trange=trange,
                   datatype='mag_RTN_4_Sa_per_Cyc',
                   time_clip=True,
                   username='none',
                   password='none',
                   version='v02')

        var_dat = pyspedas.get_data('psp_fld_l2_mag_RTN_4_Sa_per_Cyc')

        # If multiple versions are loaded in the same call, times will be non-monotonic
        t = var_dat.times
        dt = t[1:]-t[0:-1]
        dtmin = np.min(dt)
        self.assertTrue(dtmin > 0.0)


    def test_fields_sqtn_rfs_v1v1_explicit_version(self):
        trange = ['2025-03-17/00:00:00', '2025-03-18/00:00:00']

        fields_vars = pyspedas.projects.psp.fields(trange=trange,
                                                   datatype='sqtn_rfs_v1v2',
                                                   time_clip=True,version='v2.1')
        self.assertTrue('electron_density' in fields_vars)

    def test_load_dfb_dbm_dvac(self):
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='dfb_dbm_dvac', level='l2')
        self.assertTrue(data_exists('psp_fld_l2_dfb_dbm_dvac12'))

    def test_load_fld_data(self):
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_rtn', level='l2', time_clip=True)
        self.assertTrue(data_exists('psp_fld_l2_mag_RTN'))
        filtered = pyspedas.projects.psp.filter_fields('psp_fld_l2_mag_RTN', [4, 16])
        self.assertTrue(data_exists('psp_fld_l2_mag_RTN_004016'))
        filtered = pyspedas.projects.psp.filter_fields('psp_fld_l2_mag_RTN', 0)
        self.assertTrue(data_exists('psp_fld_l2_mag_RTN_000'))
        filtered = pyspedas.projects.psp.filter_fields('psp_fld_l2_mag_RTN', [4, 16], keep=True)
    
    def test_load_fld_data_prefix_none(self):
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_rtn', level='l2', time_clip=True, prefix=None)
        self.assertTrue(data_exists('psp_fld_l2_mag_RTN'))

    def test_load_fld_data_suffix_none(self):
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_rtn', level='l2', time_clip=True, suffix=None)
        self.assertTrue(data_exists('psp_fld_l2_mag_RTN'))

    def test_load_fld_data_prefix_suffix(self):
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_rtn', level='l2', time_clip=True, prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_psp_fld_l2_mag_RTN_suf'))

    def test_load_fld_1min(self):
        fields_vars = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_rtn_1min', level='l2')
        filtered = pyspedas.projects.psp.filter_fields('psp_fld_l2_mag_RTN_1min', [4, 16])
        self.assertTrue(data_exists('psp_fld_l2_mag_RTN_1min'))
        self.assertTrue(data_exists('psp_fld_l2_quality_flags'))
        notplot = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_rtn_1min', level='l2', notplot=True)
        self.assertTrue('psp_fld_l2_mag_RTN_1min' in notplot.keys())

    def test_load_fld_rtn_4_per_cyc(self):
        fields = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_rtn_4_per_cycle', level='l2')
        filtered = pyspedas.projects.psp.filter_fields('psp_fld_l2_mag_RTN_4_Sa_per_Cyc', [4, 16])
        self.assertTrue(data_exists('psp_fld_l2_mag_RTN_4_Sa_per_Cyc'))
        self.assertTrue(data_exists('psp_fld_l2_quality_flags'))

    def test_load_fld_sc_4_per_cyc(self):
        fields = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_sc_4_per_cycle',
                                     level='l2')
        filtered = pyspedas.projects.psp.filter_fields('psp_fld_l2_mag_SC_4_Sa_per_Cyc', [4, 16])
        self.assertTrue(data_exists('psp_fld_l2_mag_SC_4_Sa_per_Cyc'))
        self.assertTrue(data_exists('psp_fld_l2_quality_flags'))

    def test_load_sqtn_rfs_v1v2(self):
        fields = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='sqtn_rfs_v1v2')
        filtered = pyspedas.projects.psp.filter_fields('electron_density', [4, 16])
        self.assertTrue(data_exists('electron_density'))
        self.assertTrue(data_exists('electron_core_temperature'))

    def test_load_dfb_dc_spec(self):
        fields = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='dfb_dc_spec')
        filtered = pyspedas.projects.psp.filter_fields('psp_fld_l2_dfb_dc_spec_dV12hg', [4, 16])
        self.assertTrue(data_exists('psp_fld_l2_dfb_dc_spec_dV12hg'))
        self.assertTrue(data_exists('psp_fld_l2_dfb_dc_spec_SCMdlfhg'))

    def test_load_dfb_ac_xspec(self):
        fields = pyspedas.projects.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='dfb_ac_xspec')
        filtered = pyspedas.projects.psp.filter_fields('psp_fld_l2_dfb_ac_xspec_power_ch1_SCMdlfhg', [4, 16])
        self.assertTrue(data_exists('psp_fld_l2_dfb_ac_xspec_power_ch1_SCMdlfhg'))
        self.assertTrue(data_exists('psp_fld_l2_dfb_ac_xspec_power_ch1_SCMdlfhg'))

    def test_load_spc_data(self):
        spc_vars = pyspedas.projects.psp.spc(trange=['2018-11-5', '2018-11-6'], datatype='l3i', level='l3')
        self.assertTrue(data_exists('psp_spc_np_fit'))
        self.assertTrue(data_exists('psp_spc_np_fit_uncertainty'))
        self.assertTrue(data_exists('psp_spc_wp_fit'))
        self.assertTrue(data_exists('psp_spc_vp_fit_SC'))
        self.assertTrue(data_exists('psp_spc_vp_fit_RTN'))
        self.assertTrue(data_exists('psp_spc_np1_fit'))

    def test_load_spc_data_prefix_none(self):
        spc_vars = pyspedas.projects.psp.spc(trange=['2018-11-5', '2018-11-6'], datatype='l3i', level='l3', prefix=None)
        self.assertTrue(data_exists('psp_spc_np_fit'))

    def test_load_spc_data_suffix_none(self):
        spc_vars = pyspedas.projects.psp.spc(trange=['2018-11-5', '2018-11-6'], datatype='l3i', level='l3', suffix=None)
        self.assertTrue(data_exists('psp_spc_np_fit'))
    
    def test_load_spc_data_suffix(self):
        spc_vars = pyspedas.projects.psp.spc(trange=['2018-11-5', '2018-11-6'], datatype='l3i', level='l3', prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_psp_spc_np_fit_suf'))

    def test_load_spc_data_varnames(self):
        spc_vars = pyspedas.projects.psp.spc(trange=['2018-11-5', '2018-11-6'], datatype='l3i', level='l3', varnames=['np_fit'])
        self.assertTrue(data_exists('psp_spc_np_fit'))

    def test_load_spe_data(self):
        spe_vars = pyspedas.projects.psp.spe(trange=['2018-11-5', '2018-11-6'], datatype='spa_sf1_32e', level='l2')
        self.assertTrue(data_exists('psp_spe_EFLUX'))
        self.assertTrue(data_exists('psp_spe_QUALITY_FLAG'))
    
    def test_load_spe_data_prefix_none(self):
        spe_vars = pyspedas.projects.psp.spe(trange=['2018-11-5', '2018-11-6'], datatype='spa_sf1_32e', level='l2', prefix=None)
        self.assertTrue(data_exists('psp_spe_EFLUX'))
    
    def test_load_spe_data_suffix_none(self):
        spe_vars = pyspedas.projects.psp.spe(trange=['2018-11-5', '2018-11-6'], datatype='spa_sf1_32e', level='l2', suffix=None)
        self.assertTrue(data_exists('psp_spe_EFLUX'))
    
    def test_load_spe_data_prefix_suffix(self):
        spe_vars = pyspedas.projects.psp.spe(trange=['2018-11-5', '2018-11-6'], datatype='spa_sf1_32e', level='l2', prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_psp_spe_EFLUX_suf'))

    def test_load_spi_data(self):
        spi_vars = pyspedas.projects.psp.spi(trange=['2018-11-5', '2018-11-6'], datatype='spi_sf0a_mom_inst', level='l3')
        self.assertTrue(data_exists('psp_spi_DENS'))
        self.assertTrue(data_exists('psp_spi_VEL'))
        self.assertTrue(data_exists('psp_spi_T_TENSOR'))
        self.assertTrue(data_exists('psp_spi_TEMP'))
        self.assertTrue(data_exists('psp_spi_EFLUX_VS_ENERGY'))
        self.assertTrue(data_exists('psp_spi_EFLUX_VS_THETA'))
        self.assertTrue(data_exists('psp_spi_EFLUX_VS_PHI'))

    def test_load_spi_data_prefix_none(self):
        spi_vars = pyspedas.projects.psp.spi(trange=['2018-11-5', '2018-11-6'], datatype='spi_sf0a_mom_inst', level='l3', prefix=None)
        self.assertTrue(data_exists('psp_spi_DENS'))
    
    def test_load_spi_data_suffix_none(self):
        spi_vars = pyspedas.projects.psp.spi(trange=['2018-11-5', '2018-11-6'], datatype='spi_sf0a_mom_inst', level='l3', suffix=None)
        self.assertTrue(data_exists('psp_spi_DENS'))
    
    def test_load_spi_data_prefix_suffix(self):
        spi_vars = pyspedas.projects.psp.spi(trange=['2018-11-5', '2018-11-6'], datatype='spi_sf0a_mom_inst', level='l3', prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_psp_spi_DENS_suf'))

    def test_load_epihi_data(self):
        epihi_vars = pyspedas.projects.psp.epihi(trange=['2018-11-5', '2018-11-5/06:00'], datatype='let1_rates1h', level='l2')
        self.assertTrue(data_exists('psp_epihi_B_He_Rate'))
        self.assertTrue(data_exists('psp_epihi_R1A_He_BIN'))
        self.assertTrue(data_exists('psp_epihi_R3B_He_BIN'))
        self.assertTrue(data_exists('psp_epihi_R6A_He_BIN'))
    
    def test_load_epihi_data_prefix_none(self):
        epihi_vars = pyspedas.projects.psp.epihi(trange=['2018-11-5', '2018-11-5/06:00'], datatype='let1_rates1h', level='l2', prefix=None)
        self.assertTrue(data_exists('psp_epihi_B_He_Rate'))
    
    def test_load_epihi_data_suffix_none(self):
        epihi_vars = pyspedas.projects.psp.epihi(trange=['2018-11-5', '2018-11-5/06:00'], datatype='let1_rates1h', level='l2', suffix=None)
        self.assertTrue(data_exists('psp_epihi_B_He_Rate'))

    def test_load_epihi_data_prefix_suffix(self):
        epihi_vars = pyspedas.projects.psp.epihi(trange=['2018-11-5', '2018-11-5/06:00'], datatype='let1_rates1h', level='l2', prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_psp_epihi_B_He_Rate_suf'))

    def test_load_epi_data(self):
        epilo_vars = pyspedas.projects.psp.epi()
        self.assertTrue(data_exists('psp_isois_HET_A_Electrons_Rate_TS'))
        self.assertTrue(data_exists('psp_isois_HET_A_H_Rate_TS'))
        self.assertTrue(data_exists('psp_isois_A_H_Rate_TS'))
        self.assertTrue(data_exists('psp_isois_A_Heavy_Rate_TS'))
        self.assertTrue(data_exists('psp_isois_H_CountRate_ChanP_SP'))
        self.assertTrue(data_exists('psp_isois_Electron_CountRate_ChanE'))

    def test_load_epi_data_prefix_none(self):
        epilo_vars = pyspedas.projects.psp.epi(prefix=None)
        self.assertTrue(data_exists('psp_isois_HET_A_Electrons_Rate_TS'))

    def test_load_epi_data_suffix_none(self):
        epilo_vars = pyspedas.projects.psp.epi(suffix=None)
        self.assertTrue(data_exists('psp_isois_HET_A_Electrons_Rate_TS'))

    def test_l3_filtering(self):
        import pyspedas.projects.psp as psp
        from pyspedas import tplot, options

        # Load data during a rotation

        trange = ['2020-08-05/12:00', '2020-08-05/20:00']
        psp.fields(trange=trange, time_clip=True, datatype='rfs_hfr', level='l3')

        # Set scale to show variation during rotation

        options('psp_fld_l3_rfs_hfr_auto_averages_ch0_V1V2', 'zrange', [1e-17, 5e-17])

        # Filter out data during rotation

        psp.filter_fields('psp_fld_l3_rfs_hfr_auto_averages_ch0_V1V2', [8])

        # Plot data, filtered data, and attitude data showing rotation

        tplot(['psp_fld_l3_rfs_hfr_auto_averages_ch0_V1V2',
               'psp_fld_l3_rfs_hfr_auto_averages_ch0_V1V2_008',
               'psp_fld_l3_rfs_hfr_ch0_V1V2_J2000'],display=global_display, save_png='psp_l3filter.png')
        self.assertTrue(data_exists('psp_fld_l3_rfs_hfr_auto_averages_ch0_V1V2_008'))

    def test_load_epi_data_prefix_suffix(self):
        epilo_vars = pyspedas.projects.psp.epi(prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_psp_isois_HET_A_Electrons_Rate_TS_suf'))

    def test_downloadonly(self):
        files = pyspedas.projects.psp.epilo(downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()



import unittest
import pyspedas
import logging
from pyspedas.tplot_tools import data_exists, del_data

try:
    from viresclient import SwarmRequest
except ImportError:
    logging.info('The viresclient package is needed for this operation, but does not appear to be installed.')
    logging.info('To use this feature, install viresclient with "pip install viresclient".')
    logging.info('If pip install fails, try "conda install viresclient".')


class LoadTestCases(unittest.TestCase):
    def test_get_collections(self):
        from pyspedas.vires.config import CONFIG
        self.assertNotEqual(CONFIG['access_token'],'')
        collections = SwarmRequest(url="https://vires.services/ows",token=CONFIG['access_token']).available_collections(details=False)
        self.assertTrue('SW_OPER_MAGA_LR_1B' in collections['MAG'])
        self.assertTrue('SW_OPER_MODC_SC_1B' in collections['MOD_SC'])

    def test_load_mag_data(self):
        vires_vars = pyspedas.vires.load(trange=['2014-01-01T00:00', '2014-01-01T01:00'],
                                       collection="SW_OPER_MAGA_LR_1B",
                                       measurements=["F", "B_NEC"],
                                       models=["CHAOS-Core"],
                                       sampling_step="PT10S",
                                       auxiliaries=["QDLat", "QDLon"])
        self.assertTrue('Longitude' in vires_vars)
        self.assertTrue('Latitude' in vires_vars)
        self.assertTrue('B_NEC' in vires_vars)
        self.assertTrue(data_exists('Longitude'))
        self.assertTrue(data_exists('Latitude'))
        self.assertTrue(data_exists('B_NEC'))

    def test_load_no_trange(self):
        del_data('*')
        vires_vars = pyspedas.vires.load(collection="SW_OPER_MAGA_LR_1B",
                                       measurements=["F", "B_NEC"],
                                       models=["CHAOS-Core"],
                                       sampling_step="PT10S",
                                       auxiliaries=["QDLat", "QDLon"])
        self.assertTrue(vires_vars is None)
        self.assertFalse(data_exists('Longitude'))
        self.assertFalse(data_exists('Latitude'))
        self.assertFalse(data_exists('B_NEC'))

if __name__ == '__main__':
    unittest.main()



import os
import unittest
from pyspedas.tplot_tools import data_exists, tplot_names, del_data
from pyspedas.projects import maven
from pyspedas.projects.maven.download_files_utilities import get_orbit_files, merge_orbit_files, get_file_from_site
from pyspedas.projects.maven.maven_kp_to_tplot import maven_kp_to_tplot
from pyspedas.projects.maven.utilities import get_latest_iuvs_files_from_date_range
import time
import collections
from pyspedas.projects.maven.config import CONFIG
from datetime import datetime

# We need sleep time to avoid "HTTP Error 429: Too Many Requests"
# As of January 2025, this may no longer be necessary for MAVEN
sleep_time = 1


def get_kp_dict():
    data = maven.kp()
    local_data_dir = CONFIG["local_data_dir"]
    fn = [
        os.path.join(
            local_data_dir,
            "maven/data/sci/kp/insitu/2016/01/mvn_kp_insitu_20160101_v24_r02.tab",
        ),
        os.path.join(
            local_data_dir,
            "maven/data/sci/kp/insitu/2016/01/mvn_kp_insitu_20160102_v24_r02.tab",
        ),
    ]
    return maven_kp_to_tplot(filename=fn, notplot=True)


class OrbitTestCases(unittest.TestCase):
    def test_get_merge_orbit_files(self):
        from pyspedas.projects.maven.config import CONFIG

        get_orbit_files()
        merge_orbit_files()
        orbfilepath = os.path.join(
            CONFIG["local_data_dir"], "orbitfiles", "merged_maven_orbits.orb"
        )
        self.assertTrue(os.path.exists(orbfilepath))


class LoadTestCases(unittest.TestCase):
    def test_load_kp_data(self):
        del_data("*")
        data = maven.kp()
        self.assertTrue(data_exists("mvn_kp::spacecraft::geo_x"))
        time.sleep(sleep_time)

    def test_load_kp_spdf_data(self):
        del_data("*")
        data = maven.kp(spdf=True)
        self.assertTrue(data_exists("LPW_Electron_density"))
        time.sleep(sleep_time)

    def test_load_kp_iuvs_occ_data(self):
        del_data("*")
        data = maven.kp(trange=["2016-01-18","2016-01-19"],iuvs=True)
        self.assertTrue(data_exists("mvn_kp::spacecraft::geo_x"))
        dt1 = datetime.strptime("2016-01-18", "%Y-%m-%d")
        dt2 = datetime.strptime("2016-01-19", "%Y-%m-%d")
        fnames = get_latest_iuvs_files_from_date_range(dt1,dt2)
        self.assertTrue(len(fnames) > 0)
        self.assertTrue("mvn_kp_iuvs_occ-02533_20160118T125134_v13_r01.tab" in fnames[0])
        time.sleep(sleep_time)

    def test_load_kp_iuvs_periapse_data(self):
        del_data("*")
        data = maven.kp(trange=["2015-03-07","2015-03-08"],iuvs=True)
        self.assertTrue(data_exists("mvn_kp::spacecraft::geo_x"))
        dt1 = datetime.strptime("2015-03-07", "%Y-%m-%d")
        dt2 = datetime.strptime("2015-03-08", "%Y-%m-%d")
        # fnames = get_latest_iuvs_files_from_date_range(dt1,dt2)
        # print(fnames)
        #self.assertTrue(len(fnames) > 0)
        #self.assertTrue("mvn_kp_iuvs_00850_20150308T221253_v13_r01.tab" in fnames[0])
        time.sleep(sleep_time)

    def test_load_kp_iuvs_corona_data(self):
        del_data("*")
        data = maven.kp(trange=["2016-01-14","2016-01-15"],iuvs=True)
        self.assertTrue(data_exists("mvn_kp::spacecraft::geo_x"))
        dt1 = datetime.strptime("2016-01-07", "%Y-%m-%d")
        dt2 = datetime.strptime("2016-01-08", "%Y-%m-%d")
        # fnames = get_latest_iuvs_files_from_date_range(dt1,dt2)
        # print(fnames)
        #self.assertTrue(len(fnames) > 0)
        #self.assertTrue("mvn_kp_iuvs_00850_20150308T221253_v13_r01.tab" in fnames[0])
        time.sleep(sleep_time)

    def test_kp_param_errors(self):
        from pyspedas.projects.maven.kp_utilities import param_list, param_range, range_select
        # bad value in kp dict
        kp = {}
        kp["foo"] = "bar"
        with self.assertLogs(level="WARNING") as log:
            param_list = param_list(kp)
            self.assertTrue("unexpected value type" in log.output[0])
        kp_insitu = {}
        kp_iuvs = {}
        kp_insitu["TimeString"] = ["1970-01-01", "1970-01-02"]
        kp_insitu["Orbit"] = [0,1]
        param_range(kp_insitu)

        kp_iuvs["TimeString"] = ["1971-01-01", "1971-01-02"]
        kp_iuvs["Orbit"] = [10, 11]
        with self.assertLogs(level="WARNING") as log:
            param_range(kp_insitu, kp_iuvs)
            self.assertTrue("No overlap" in log.output[0])

        with self.assertLogs(level="WARNING") as log:
            range_select(kp_iuvs)
            self.assertTrue("*****ERROR*****" in log.output[0])
            i = len(log.output)
            range_select(kp_iuvs,parameter=0)
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            range_select(kp_iuvs,time=["1970-01-01"])
            i = len(log.output)
            range_select(kp_iuvs, time=["1970-01-01", 0])
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)

    def test_kp_utilities(self):
        from pyspedas.projects.maven.kp_utilities import (
            param_list,
            param_range,
            range_select,
            get_inst_obs_labels,
        )
        from pyspedas.projects.maven.kp_utilities import find_param_from_index

        kp = get_kp_dict()
        self.assertTrue(type(kp) is collections.OrderedDict)

        param_list = param_list(kp)
        self.assertTrue(len(param_list) > 0)
        print(param_list)

        param_range = param_range(kp)
        result = range_select(kp, [2440, 2445], [5], [1e9], [-1e9])
        self.assertTrue(len(result) > 0)
        result = range_select(kp, ["2016-01-01 00:00:00","2016-01-02 00:00:00"], [5], [1e9], [-1e9])
        self.assertTrue(len(result) > 0)
        labels = get_inst_obs_labels(kp, "LPW.EWAVE_LOW_FREQ")
        self.assertTrue("LPW" in labels)
        self.assertTrue("EWAVE_LOW_FREQ" in labels)
        param = find_param_from_index(kp, 5)
        self.assertTrue(param == "LPW.ELECTRON_DENSITY")
        # no min
        result = range_select(kp, ["2016-01-01 00:00:00","2016-01-02 00:00:00"], parameter=[5], maximum=[1e9])
        self.assertTrue(len(result) > 0)
        # no max
        result = range_select(kp, ["2016-01-01 00:00:00","2016-01-02 00:00:00"], parameter=[5], minimum=[1e9])
        self.assertTrue(len(result) > 0)
        # no time, param as list with int
        result = range_select(kp, parameter=[5], minimum=[-1e9], maximum=[1e9])
        # no time, param as list with int
        result = range_select(kp, parameter=[5], minimum=[-1e9], maximum=[1e9])
        self.assertTrue(len(result) > 0)
        # no time, param as list with string
        result = range_select(kp, parameter=["LPW.ELECTRON_DENSITY"], minimum=[-1e9], maximum=[1e9])
        self.assertTrue(len(result) > 0)
        # no time, param as list with string and int
        result = range_select(kp, parameter=["LPW.ELECTRON_DENSITY",6], minimum=[-1e9, -1e9], maximum=[1e9,1e9])
        self.assertTrue(len(result) > 0)
        # no time, parameter as scalar int
        result = range_select(kp, parameter=5, minimum=[-1e9], maximum=[1e9])
        self.assertTrue(len(result) > 0)
        # no time, parameter as scalar string
        result = range_select(kp, parameter="LPW.ELECTRON_DENSITY", minimum=[-1e9], maximum=[1e9])
        self.assertTrue(len(result) > 0)
        # no time, parameter, min, max as scalars
        result = range_select(kp, parameter="LPW.ELECTRON_DENSITY", minimum=-1e9, maximum=1e9)
        self.assertTrue(len(result) > 0)
        # no time, scalar param, minimum=None
        result = range_select(kp, parameter=5, maximum=[1e9])
        self.assertTrue(len(result) > 0)
        # no time, scalar param, maximum=None
        result = range_select(kp, parameter=5, maximum=[1e9])
        self.assertTrue(len(result) > 0)
        # times, scalar param, scalar max/min
        result = range_select(kp, time=[2440,2441], parameter=5, minimum=-1e9, maximum=[1e9])
        self.assertTrue(len(result) > 0)
        # times, scalar param, no min
        result = range_select(kp, time=[2440, 2441], parameter=5, maximum=[1e9])
        # times, scalar param, no max
        result = range_select(kp, time=[2440, 2441], parameter=5, minimum = [1e9])

        with self.assertLogs(level="WARNING") as log:
            # mismatched time types
            result = range_select(kp, [2440, "2020/04/01"], [5], [1e9], [-1e9])
            self.assertTrue("*****WARNING*****" in log.output[0])
            i = len(log.output)
            # only one time
            result = range_select(kp, [2440], [5], [1e9], [-1e9])
            self.assertTrue("*****WARNING*****" in log.output[i])
            i = len(log.output)
            # parameter but no max/min
            result = range_select(kp, ["2016-01-01 00:00:00","2016-01-02 00:00:00"], parameter=[5])
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # len(min) doesn't match param
            result = range_select(kp, ["2016-01-01 00:00:00","2016-01-02 00:00:00"], parameter=[5], minimum=[-1e9,-1e9], maximum=[1e9])
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # len(max) doesn't match param
            result = range_select(kp, ["2016-01-01 00:00:00","2016-01-02 00:00:00"], parameter=[5], minimum=[-1e9], maximum=[1e9, 1e9])
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            self.assertTrue(len(result) > 0)
            # len(min) doesn't match param, no max
            result = range_select(kp, ["2016-01-01 00:00:00","2016-01-02 00:00:00"], parameter=[5], minimum=[-1e9,-1e9])
            self.assertTrue(len(result) > 0)
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # len(max) doesn't match param, no min
            result = range_select(kp, ["2016-01-01 00:00:00","2016-01-02 00:00:00"], parameter=[5], maximum=[1e9, 1e9])
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # no time, param as list with string, int, and float
            result = range_select(kp, parameter=["LPW.ELECTRON_DENSITY", 6, 1.2], minimum=[-1e9, -1e9, -1e9], maximum=[1e9, 1e9, 1e9])
            self.assertTrue(len(result) > 0)
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # no time, param as scalar float
            result = range_select(kp, parameter=1.2, minimum=[-1e9], maximum=[1e9])
            self.assertTrue(len(result) > 0)
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # no time, no param
            result = range_select(kp, minimum=[-1e9], maximum=[1e9])
            self.assertTrue(len(result) > 0)
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # no time, scalar param, no bounds
            result = range_select(kp, parameter=5)
            self.assertTrue(len(result) > 0)
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # no time, len(min) doesn't match param
            result = range_select(kp, parameter=[5], minimum=[-1e9,-1e9], maximum=[1e9])
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # no time, len(max) doesn't match param
            result = range_select(kp,  parameter=[5], minimum=[-1e9], maximum=[1e9, 1e9])
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            self.assertTrue(len(result) > 0)
            # no time, len(min) doesn't match param, no max
            result = range_select(kp,  parameter=[5], minimum=[-1e9,-1e9])
            self.assertTrue(len(result) > 0)
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # no time, len(max) doesn't match param, no min
            result = range_select(kp, parameter=[5], maximum=[1e9, 1e9])
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # only one time, no param
            result = range_select(kp, [2440])
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # Malformed times, no parameter
            result = range_select(kp, [{}, {}])
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # Malformed times, parameter given
            result = range_select(kp, [{}, {}], parameter=[5], minimum=[-1e9], maximum=[1e9])
            self.assertTrue("*****WARNING*****" in log.output[i])
            i = len(log.output)
            # times, scalar param, len(min) doesn't match param
            result = range_select(kp, time=[2440, 2441], parameter=5, minimum=[-1e9, -1e9], maximum=[1e9])
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # times, scalar param, len(max) doesn't match param
            result = range_select(kp, time=[2440, 2441], parameter=5, minimum=[-1e9], maximum=[1e9, 1e9])
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # times, scalar param, no bounds
            result = range_select(kp, time=[2440, 2441], parameter=5)
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # invalid parameter value
            a,b = get_inst_obs_labels(kp, "foo")
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # too many components
            a,b = get_inst_obs_labels(kp, "foo.bar.baz")
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)
            # Invalid numeric index
            ind = find_param_from_index(kp,999)
            self.assertTrue("*****ERROR*****" in log.output[i])
            i = len(log.output)

    def test_get_file_from_site_private(self):
        f = 'mvn_mag_l2_2016002ss1s_20160102_v01_r01.xml'
        public=False
        full_path='maven_data/maven/data/sci/mag/l2/2016/01'
        # We don't have credentials to the private site yet, so this will fail.
        try:
            get_file_from_site(f, public, full_path)
        except Exception as e:
            pass

    def test_load_mag_data(self):
        from pyspedas.projects.maven.utilities import get_l2_files_from_date

        del_data("*")
        data = maven.mag(datatype="ss1s")
        self.assertTrue(len(tplot_names("OB_B*"))>0)
        dt = datetime.strptime("2016-01-01/12:00:00", "%Y-%m-%d/%H:%M:%S")
        files = get_l2_files_from_date(dt, "mag")
        self.assertTrue(len(files) > 0)
        time.sleep(sleep_time)

    def test_load_mag_data_private(self):
        from pyspedas.projects.maven.utilities import get_l2_files_from_date

        del_data("*")
        # We don't have credentials to the private site yet, so this is expected to fail
        try:
            data = maven.mag(datatype="ss1s", public=False)
        except Exception as e:
            pass

    def test_load_mag_byorbit_data(self):
        del_data("*")
        data = maven.mag(trange=[500, 501], datatype="ss1s")
        self.assertTrue(len(tplot_names("OB_B*"))>0)
        time.sleep(sleep_time)

    def test_load_sta_data(self):
        del_data("*")
        # No datatype means "load everything"
        data = maven.sta()
        self.assertTrue(data_exists("hkp_raw_2a-hkp"))
        self.assertTrue(data_exists("hkp_2a-hkp"))
        self.assertTrue(data_exists("data_d0-32e4d16a8m"))
        self.assertTrue(data_exists("theta_d1-32e4d16a8m"))
        time.sleep(sleep_time)

    def test_load_sta_hkp_data(self):
        del_data("*")
        data = maven.sta(datatype=["2a"])
        self.assertTrue(data_exists("hkp_raw_2a-hkp"))
        self.assertTrue(data_exists("hkp_2a-hkp"))
        time.sleep(sleep_time)

    def test_load_swea_data(self):
        del_data("*")
        data = maven.swea()
        self.assertTrue(data_exists("diff_en_fluxes_svyspec"))
        time.sleep(sleep_time)

    def test_load_swia_data(self):
        del_data("*")
        data = maven.swia()
        self.assertTrue(data_exists("spectra_diff_en_fluxes_onboardsvyspec"))
        time.sleep(sleep_time)

    def test_load_sep_data(self):
        del_data("*")
        data = maven.sep()
        self.assertTrue(data_exists("f_ion_flux_tot_s2-cal-svy-full"))
        time.sleep(sleep_time)

    def test_load_lpw_data(self):
        del_data("*")
        data = maven.lpw()
        self.assertTrue(data_exists("mvn_lpw_lp_iv_l2_lpiv"))
        time.sleep(sleep_time)

    def test_load_euv_data(self):
        del_data("*")
        data = maven.euv()
        self.assertTrue(data_exists("mvn_euv_calib_bands_bands"))
        time.sleep(sleep_time)

    @unittest.skip("Unable to load RSE data in *.tab file format")
    def test_load_rse_data(self):
        del_data('*')
        data = maven.rse()
        tplot_names()
        self.assertTrue(data_exists('mvn_kp::spacecraft::geo_x'))
        print(data)
        time.sleep(sleep_time)

    def test_load_iuv_data(self):
        del_data("*")
        data = maven.iuv()
        self.assertTrue(data_exists("mvn_kp::spacecraft::altitude"))
        time.sleep(sleep_time)

    def test_load_ngi_data(self):
        del_data("*")
        data = maven.ngi()
        self.assertTrue(data_exists("mvn_kp::spacecraft::altitude"))
        time.sleep(sleep_time)


if __name__ == "__main__":
    unittest.main()



import logging
import numpy as np
from scipy.stats import linregress


def pwrspc(time, quantity, noline=False, nohanning=False, bin=3, notperhz=False):
    """
    Compute the power spectrum of a given time series.

    Parameters
    ----------
        time (array):
            The time array.
        quantity (array):
            The data array for which the power spectrum is to be computed.
            Should be one dimensional and the same length as time.
        noline (bool):
            If True, straight line is not subtracted from the data.
        nohanning (bool):
            If True, no Hanning window is applied to the data.
        bin (int):
            Bin size for binning the data. Default is 3.
        notperhz (bool):
            If True, the output units are the square of the input units.

    Returns
    -------
    tuple
        Tuple containing:
            - freq (array):
                The frequency array.
            - power (array):
                The computed power spectrum.

    Notes:
        This is similar to IDL pwrspc.pro routine.

        A Hanning window is applied to the input data, and its power is divided out of the returned spectrum.
        A straight line is subtracted from the data to reduce spurious power due to sawtooth behavior of a background.
        Units are (units)^2 where units are the units of the input quantity. Frequency is in 1/time units.
        Thus, the output represents the mean squared amplitude of the signal at each specific frequency.
        The total (sum) power under the curve is equal to the mean (over time) power of the oscillation in the time domain.
        If the keyword notperhz is True, then power is in units^2. If notperhz is False (default), power is in units^2/Hz.


    Examples
    --------

        >>> # Compute the power spectrum of a given time series
        >>> from pyspedas import pwrspc
        >>> time = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        >>> quantity = [1,2,3,1,2,3,1,2,3,1]
        >>> freq, power = pwrspc(time, quantity)
        >>> print(freq, power)
    """

    t = np.array(time, dtype=np.float64)
    x = np.array(quantity, dtype=np.float64)

    # If the dimensions of the input arrays are not the same, and not one dimension, return
    if t.ndim != 1 or x.ndim != 1 or len(t) != len(x) or len(t) < 1:
        logging.error(
            "Both input arrays should be one dimensional and of the same length."
        )
        return np.array(None), np.array(None)

    # Subtract first point from time array
    t -= t[0]

    # Subtract straight line from data
    if not noline:
        slope, intercept, _, _, _ = linregress(t, x)
        x -= slope * t + intercept

    binsize = bin
    window = 0.0
    if not nohanning:
        window = np.hanning(len(x))
        x *= window

    nt = len(t)
    if nt % 2 != 0:
        logging.info("Needs an even number of data points, dropping last point...")
        t = t[:-1]
        x = x[:-1]
        nt -= 1

    xs2 = np.abs(np.fft.fft(x)) ** 2
    dbign = float(nt)
    logging.info("bign=" + str(dbign))

    k = np.arange(0, dbign // 2 + 1)
    tres = float(np.median(np.diff(t)))
    fk = k / (dbign * tres)

    pwr = np.zeros(nt // 2 + 1)
    pwr[0] = xs2[0] / dbign**2
    pwr[1 : nt // 2] = (xs2[1 : nt // 2] + xs2[nt : nt // 2 : -1]) / dbign**2
    pwr[-1] = xs2[-1] / dbign**2

    if not nohanning:
        wss = dbign * np.sum(window**2)
        pwr = pwr * dbign**2 / wss

    dfreq = binsize * (fk[1] - fk[0])
    npwr = len(pwr) - 1
    nfinal = int(npwr / binsize)
    iarray = np.arange(nfinal)
    power = np.zeros(nfinal)

    idx = (iarray + 0.5) * binsize + 1
    freq = [fk[int(i)] for i in idx]

    for i in range(binsize):
        power += pwr[iarray * binsize + i + 1]

    if not notperhz:
        power /= dfreq

    logging.info("dfreq=" + str(dfreq))

    return np.array(freq), np.array(power)



import pyspedas
import logging

def rotmat_get_coords(varname:str):
    """ Get input and output coordinates on a rotation matrix

    Parameters
    ----------
    varname:str
        Variable name of the rotation matrix

    Returns
    -------
    (str, str)
    Tuple containing the input and output coordinates (or None if not set)

    """

    md=pyspedas.get_data(varname, metadata=True)
    data_att = md.get('data_att')
    if data_att is None:
        logging.warning(f"Rotation matrix {varname} has no input or output coordinates set")
        return (None, None)
    in_coords = data_att.get('input_coord_sys')
    if in_coords is None:
        logging.warning(f"Rotation matrix {varname} has no input coordinates set")
    out_coords = data_att.get('output_coord_sys')
    if out_coords is None:
        logging.warning(f"Rotation matrix {varname} has no output coordinates set")
    return (in_coords, out_coords)



"""
Python implementation of IDL reduce_tres function.

Uses pyspedas rebin function.
"""
import numpy as np
from pyspedas.analysis.rebin import rebin


def reduce_tres(dat, n):
    """
    Rebin arrays by a specified factor.

    Python translation of IDL reduce_tres.pro function.

    Parameters
    ----------
    dat : array_like
        Input data array to be rebinned
    n : int
        Rebinning factor. If n <= 1, returns original data unchanged.

    Returns
    -------
    array_like
        Rebinned array with reduced time resolution, or original data if n <= 1
    """
    if n <= 1:
        return dat

    dat = np.asarray(dat)
    dim = dat.shape

    # Calculate how many elements to truncate to make divisible by n
    m = dim[0] % n
    l = dim[0] - m - 1

    # Handle different dimensionalities
    if dat.ndim == 1:
        # IDL: return,rebin(dat[0:l],dim[0]/n)
        truncated = dat[: l + 1]
        new_dim = len(truncated) // n
        return rebin(truncated, new_dim)

    elif dat.ndim == 2:
        # IDL: return,rebin(dat[0:l,*],dim[0]/n,dim[1])
        truncated = dat[: l + 1, :]
        new_dim0 = truncated.shape[0] // n
        new_dim1 = dim[1]
        return rebin(truncated, (new_dim0, new_dim1))

    elif dat.ndim == 3:
        # IDL: return,rebin(dat[0:l,*,*],dim[0]/n,dim[1],dim[2])
        truncated = dat[: l + 1, :, :]
        new_dim0 = truncated.shape[0] // n
        new_dim1 = dim[1]
        new_dim2 = dim[2]
        return rebin(truncated, (new_dim0, new_dim1, new_dim2))

    else:
        # For higher dimensions, return 0 (matching IDL behavior)
        return 0



import pyspedas
from pyspedas.tplot_tools import get_y_range
from copy import deepcopy
import logging


def replace_metadata(tplot_name, new_metadata):
    """
    This function will replace all the metadata in a tplot variable.

    Parameters
    ----------
    tplot_name : str
        The name of the tplot variable.
    new_metadata : dict
        A dictionary with metadata values. A deep copy will be performed so that
        no references to new_metadata are retained.

    Returns
    -------
    None

    Examples
    --------
    >>> # Copy Variable 1 metadata into Variable 2, which must already exist
    >>> import pyspedas
    >>> meta = pyspedas.get_data('Variable1', metadata=True)
    >>> pyspedas.replace_metadata("Variable2", meta)
    """

    # if old name input is a number, convert to corresponding name
    if isinstance(tplot_name, int):
        tplot_name = pyspedas.tplot_tools.data_quants[tplot_name].name

    # check if old name is in current dictionary
    if tplot_name not in pyspedas.tplot_tools.data_quants.keys():
        logging.error(f"replace_metadata: {tplot_name} is currently not in pyspedas")
        return

    original_create_time = pyspedas.tplot_tools.data_quants[tplot_name].attrs['plot_options']['create_time']
    original_error = pyspedas.tplot_tools.data_quants[tplot_name].attrs['plot_options']['error']
    original_trange = pyspedas.tplot_tools.data_quants[tplot_name].attrs['plot_options']['trange']
    xaxis_opt = pyspedas.tplot_tools.data_quants[tplot_name].attrs['plot_options']['xaxis_opt']
    yaxis_opt = pyspedas.tplot_tools.data_quants[tplot_name].attrs['plot_options']['yaxis_opt']
    zaxis_opt = pyspedas.tplot_tools.data_quants[tplot_name].attrs['plot_options']['zaxis_opt']
    xaxis_opt['crosshair'] = 'X'
    yaxis_opt['crosshair'] = 'Y'
    zaxis_opt['crosshair'] = 'Z'
    xaxis_opt['x_axis_type'] = 'linear'
    yaxis_opt['y_axis_type'] = 'linear'
    zaxis_opt['z_axis_type'] = 'linear'
    line_opt = {}
    time_bar = []
    extras = dict(panel_size=1, border=True)
    links = {}

    md_copy = deepcopy(new_metadata)
    
    if 'plot_options' not in md_copy.keys():
        md_copy['plot_options'] = {}
        md_copy['plot_options']['xaxis_opt'] = xaxis_opt
        md_copy['plot_options']['yaxis_opt'] = yaxis_opt
        md_copy['plot_options']['zaxis_opt'] = zaxis_opt
        md_copy['plot_options']['line_opt'] = line_opt
        md_copy['plot_options']['trange'] = original_trange
        md_copy['plot_options']['time_bar'] = time_bar
        md_copy['plot_options']['extras'] = extras
        md_copy['plot_options']['create_time'] = original_create_time
        md_copy['plot_options']['links'] = links
        #md_copy['plot_options']['spec_bins_ascending'] = _check_spec_bins_ordering(times, spec_bins)
        md_copy['plot_options']['overplots'] = []
        md_copy['plot_options']['overplots_mpl'] = []
        md_copy['plot_options']['interactive_xaxis_opt'] = {}
        md_copy['plot_options']['interactive_yaxis_opt'] = {}
        md_copy['plot_options']['error'] = original_error

    
    pyspedas.tplot_tools.data_quants[tplot_name].attrs = md_copy
    pyspedas.tplot_tools.data_quants[tplot_name].attrs["plot_options"]["yaxis_opt"]["y_range"] = (
        get_y_range(pyspedas.tplot_tools.data_quants[tplot_name])
    )

    return




import os
import unittest
from pyspedas.tplot_tools import data_exists, del_data, tplot_names

import pyspedas

class LoadTestCases(unittest.TestCase):
    def test_camera_omti_asi_data(self):
        del_data()
        omti_vars = pyspedas.projects.erg.camera_omti_asi(site='ath', trange=['2020-01-20', '2020-01-21'])
        self.assertTrue(data_exists('omti_asi_ath_5577_image_raw'))
        self.assertTrue('omti_asi_ath_5577_image_raw' in omti_vars)

    def test_load_isee_fluxgate_data(self):
        del_data()
        fg_vars = pyspedas.projects.erg.gmag_isee_fluxgate(trange=['2020-08-01/00:00:00','2020-08-01/02:00:00'], site='lcl')
        self.assertTrue(data_exists('isee_fluxgate_mag_lcl_64hz_hdz'))
        self.assertTrue('isee_fluxgate_mag_lcl_64hz_hdz' in fg_vars)

    def test_load_stel_fluxgate_data(self):
        del_data()
        fg_vars = pyspedas.projects.erg.gmag_stel_fluxgate(trange=['2020-08-01/00:00:00','2020-08-01/02:00:00'], site='lcl')
        self.assertTrue(data_exists('isee_fluxgate_mag_lcl_64hz_hdz'))
        self.assertTrue('isee_fluxgate_mag_lcl_64hz_hdz' in fg_vars)

    def test_load_isee_induction_data(self):
        del_data()
        ind_vars = pyspedas.projects.erg.gmag_isee_induction(trange=['2020-08-01/00:00:00', '2020-08-01/02:00:00'], site='msr')
        self.assertTrue(data_exists('isee_induction_db_dt_msr'))
        self.assertTrue('isee_induction_db_dt_msr' in ind_vars)

    def test_load_isee_induction_frequency_dependent_data(self):
        del_data()
        ind_vars = pyspedas.projects.erg.gmag_isee_induction(trange=['2020-08-01/00:00:00', '2020-08-01/02:00:00'], site='msr', frequency_dependent=True)
        self.assertTrue(type(ind_vars) == dict)
        self.assertTrue('msr' in ind_vars)

    def test_load_stel_induction_data(self):
        del_data()
        ind_vars = pyspedas.projects.erg.gmag_stel_induction(trange=['2020-08-01/00:00:00', '2020-08-01/02:00:00'], site='msr')
        self.assertTrue(data_exists('isee_induction_db_dt_msr'))
        self.assertTrue('isee_induction_db_dt_msr' in ind_vars)

    def test_load_magdas_1sec_data(self):
        del_data()
        magdas_vars = pyspedas.projects.erg.gmag_magdas_1sec(trange=["2010-01-01/00:00:00", "2010-01-01:02:00:00"], site='ama')
        self.assertTrue(data_exists('magdas_mag_ama_1sec_hdz'))
        self.assertTrue('magdas_mag_ama_1sec_hdz' in magdas_vars)

    def test_load_mm210_data(self):
        del_data()
        mm210_vars = pyspedas.projects.erg.gmag_mm210(trange=["2005-01-01", "2005-01-02"], site='adl', datatype='1min')
        self.assertTrue(data_exists('mm210_mag_adl_1min_hdz'))
        self.assertTrue('mm210_mag_adl_1min_hdz' in mm210_vars)

    def test_load_sd_fit_data(self):
        del_data()
        sd_vars = pyspedas.projects.erg.sd_fit(trange=['2018-10-14/00:00:00', '2018-10-14/02:00:00'], site='ade')
        self.assertTrue(data_exists('sd_ade_azim_no_5'))
        self.assertTrue('sd_ade_azim_no_5' in sd_vars)

    def test_load_isee_brio_data(self):
        del_data()
        brio_vars = pyspedas.projects.erg.isee_brio(trange=['2020-08-01', '2020-08-02'], site='ath')
        self.assertTrue(data_exists('isee_brio30_ath_64hz_cna'))
        self.assertTrue('isee_brio30_ath_64hz_cna' in brio_vars)

    def test_load_isee_vlf_data(self):
        del_data()
        vlf_vars = pyspedas.projects.erg.isee_vlf(trange=['2017-03-30/12:00:00', '2017-03-30/15:00:00'], site='ath')
        self.assertTrue(data_exists('isee_vlf_ath_ch1'))
        self.assertTrue('isee_vlf_ath_ch1' in vlf_vars)

    def test_load_isee_vlf_cal_gain_data(self):
        del_data()
        vlf_vars = pyspedas.projects.erg.isee_vlf(trange=['2017-03-30/12:00:00', '2017-03-30/15:00:00'], site='ath',  cal_gain=True)
        self.assertTrue(data_exists('isee_vlf_ath_ch1'))
        self.assertTrue('isee_vlf_ath_ch1' in vlf_vars)


if __name__ == '__main__':
    unittest.main()


import os
import time
import requests
import unittest
import subprocess

from pyspedas.utilities.download import download

class DownloadTestCases(unittest.TestCase):
    """
    Cloud Awareness Unit Tests

    Depends upon moto[server] package. Install via:
        pip install moto[server]

    These tests essentially create a local mock-AWS server as a background
    process at port 3000.
    Note: The environment variables are used as mock credentials in order
          to avoid having to pass the endpoint url to fsspec calls.
    """

    @classmethod
    def setUpClass(cls):
        # Start the moto server for S3 in the background
        # https://github.com/getmoto/moto/issues/4418
        cls.moto_server = subprocess.Popen(
                ["moto_server", "-p3000"],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        # Allow the server to start properly
        time.sleep(2)

        # Set up mock AWS environment variables (fake credentials)
        os.environ["AWS_ACCESS_KEY_ID"] = "test"
        os.environ["AWS_SECRET_ACCESS_KEY"] = "test"
        os.environ["AWS_DEFAULT_REGION"] = "us-east-1"

        # Set environment to use the local Moto S3 server
        # S3 ENDPOINT for fsspec
        # ENDPOINT URL for cdflib's boto3
        os.environ["AWS_S3_ENDPOINT"] = "http://localhost:3000"
        os.environ["AWS_ENDPOINT_URL"] = "http://localhost:3000"

        # Create a bucket using direct HTTP requests
        bucket_name = "test-bucket"
        response = requests.put(f"http://localhost:3000/{bucket_name}")
        assert response.status_code == 200, "Bucket creation failed"

    @classmethod
    def tearDownClass(cls):
        # Terminate the moto server after tests
        cls.moto_server.terminate()
        cls.moto_server.communicate()

    #==========================================================================
    # Adapted unit tests (from download_tests.py) for AWS-specific URI testing.
    def test_local_path(self):
        # Remote/AWS details
        bucket_name = "test-bucket"
        s3_url = f"s3://{bucket_name}"

        # Include mock AWS credentials and endpoints
        os.environ["AWS_ACCESS_KEY_ID"] = "test"
        os.environ["AWS_SECRET_ACCESS_KEY"] = "test"
        os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
        os.environ["AWS_S3_ENDPOINT"] = "http://localhost:3000"
        os.environ["AWS_ENDPOINT_URL"] = "http://localhost:3000"

        # specifying local_path changes the local data directory
        files = download(
                local_path=s3_url + "/psp_data/spc/13",
                remote_file="https://spdf.gsfc.nasa.gov/pub/data/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf"
        )

        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == s3_url + "/psp_data/spc/13/psp_swp_spc_l3i_20190401_v01.cdf")

    def test_remote_path(self):
        # Remote/AWS details
        bucket_name = "test-bucket"
        s3_url = f"s3://{bucket_name}"

        # Include mock AWS credentials and endpoints
        os.environ["AWS_ACCESS_KEY_ID"] = "test"
        os.environ["AWS_SECRET_ACCESS_KEY"] = "test"
        os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
        os.environ["AWS_S3_ENDPOINT"] = "http://localhost:3000"
        os.environ["AWS_ENDPOINT_URL"] = "http://localhost:3000"

        # download (if not already downloaded from prior tests)
        files = download(
                local_path=s3_url + "/psp/sweap/spc/l3/l3i/2019",
                remote_file="https://spdf.gsfc.nasa.gov/pub/data/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf"
        )
        self.assertTrue(len(files) == 1)
        self.assertTrue(files[0] == s3_url + "/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf")

        # stream from remote
        with self.assertLogs(level='INFO') as log:
            files = download(
                    remote_path=s3_url,
                    remote_file="/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf"
            )
            self.assertIn("Streaming from remote", log.output[0])

    def test_force_download(self):
        # Remote/AWS details
        bucket_name = "test-bucket"
        s3_url = f"s3://{bucket_name}"

        # Include mock AWS credentials and endpoints
        os.environ["AWS_ACCESS_KEY_ID"] = "test"
        os.environ["AWS_SECRET_ACCESS_KEY"] = "test"
        os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
        os.environ["AWS_S3_ENDPOINT"] = "http://localhost:3000"
        os.environ["AWS_ENDPOINT_URL"] = "http://localhost:3000"

        files = download(
                local_path=s3_url + "/themis/tha/l1/state/2007",
                remote_file="https://spdf.gsfc.nasa.gov/pub/data/themis/tha/l1/state/2007/tha_l1_state_20070217_v01.cdf"
        )
        self.assertTrue(len(files) == 1)
        print(files)
        self.assertTrue(files[0] == s3_url + "/themis/tha/l1/state/2007/tha_l1_state_20070217_v01.cdf")

        # Download the same file without force_download, should not re-download
        with self.assertLogs(level='INFO') as log:
            files = download(
                    local_path=s3_url + "/themis/tha/l1/state/2007",
                    remote_file="https://spdf.gsfc.nasa.gov/pub/data/themis/tha/l1/state/2007/tha_l1_state_20070217_v01.cdf"
            )
            self.assertIn("File is current", log.output[0])

        # Download the same file with force_download, should re-download
        with self.assertLogs(level='INFO') as log:
            files = download(
                    local_path=s3_url + "/psp_data/spc/13",
                    remote_file="https://spdf.gsfc.nasa.gov/pub/data/psp/sweap/spc/l3/l3i/2019/psp_swp_spc_l3i_20190401_v01.cdf",
                    force_download=True
            )
            self.assertIn("Downloading", log.output[0])

if __name__ == '__main__':
    unittest.main(verbosity=2)



import logging
import re
import numpy as np



def param_list(kp):
    """
    Return a listing of all parameters present in the given
    insitu data dictionary/structure.

    Parameters:
        kp: insitu kp data structure/dictionary read from file(s)
    Output:
        ParamList: a list of all contained items and their indices.
    """
    import pandas as pd

    index = 1
    param_list_ = []
    for base_tag in kp.keys():
        if isinstance(kp[base_tag], pd.DataFrame):
            for obs_tag in kp[base_tag].columns:
                param_list_.append("#%3d %s.%s" % (index, base_tag, obs_tag))
                index += 1
        elif isinstance(kp[base_tag], pd.Series):
            param_list_.append("#%3d %s" % (index, base_tag))
            index += 1
        elif isinstance(kp[base_tag], pd.Index):
            param_list_.append("#%3d %s" % (index, base_tag))
            index += 1
        else:
            logging.warning("Warning: unexpected value type %s for tag %s",str(type(kp[base_tag])), base_tag)

    return param_list_


def param_range(kp, iuvs=None):
    """
    Print the range of times and orbits for the provided insitu data.
    If iuvs data are also provided, return only orbit numbers for IUVS data.

    Caveats:
        At present, not configured to handle (real) IUVS data.
        Current configuration of procedure assumes IUVS has identical
            time information as in-situ.

    Parameters:
        kp: insitu kp data structure/dictionary
        iuvs: IUVS kp data strucure/dictionary
    Output:
        None: prints information to screen
    """

    # First, the case where insitu data are provided
    logging.info("The loaded insitu KP data set contains data between %s and %s (orbits %d and %d)",
                 np.array(kp["TimeString"])[0],
                       np.array(kp["TimeString"])[-1],
                       np.array(kp["Orbit"])[0],
                       np.array(kp["Orbit"])[-1])

    #  Finally, the case where both insitu and IUVS are provided
    if iuvs is not None:
        logging.info("The loaded IUVS KP data set contains data between %s and %s (orbits %d and %d)",
                     np.array(iuvs["TimeString"])[0],
                     np.array(iuvs["TimeString"])[-1],
                     np.array(iuvs["Orbit"])[0],
                     np.array(iuvs["Orbit"])[-1])

        insitu_min, insitu_max = (np.nanmin([kp["Orbit"]]), np.nanmax([kp["Orbit"]]))
        if (
            np.nanmax([iuvs["Orbit"]]) < insitu_min
            or np.nanmin([iuvs["Orbit"]]) > insitu_max
        ):
            logging.warning("Warning: No overlap between supplied insitu and IUVS data structures.")
    return  # No information to return


def range_select(kp, time=None, parameter=None, maximum=None, minimum=None):
    """
    Returns a subset of the input data based on the provided time
    and/or parameter criteria.  If neither Time nor Parameter filter
    information is provided, then no subselection of data will occur.
    Any parameter used as a filtering criterion must be paired with
    either a maximum and/or a minimum value.  Open ended bounds must
    be indicated with either a value of 'None' or an empty string ('').

    Parameters:
        kp: insitu kp data structure/dictionary read from file(s)
        time: two-element time range must be either strings of format
            'yyyy-mm-ddThh:mm:ss' or integers (orbit numbers)
        parameter: Element of provided data structure/dictionary by
            which to filter data.  Parameter(s) must be either integer
            type (search by index) or string type (search by instrument
            name and observation type).  If multiple Parameters are used
            to filter the data, they must be provided as a list (mixing
            data types within a list is permitted).
        maximum: maximum value of Parameter on which to filter.  A value of
            None or '' will leave the Parameter filter unbounded above.
            The number of elements of Maximum *MUST* equal the number of
            elements of Parameter.
        minimum: minimum value of Parameter on which to filter.  A value of
            None or '' will leave the Parameter filter unbounded below.
            The number of elements of Minimum *MUST* equal the number of
            elements of Parameter.
    Output: a dictionary/structure containing the same elements as the provided
        one, but filtered according to the Time and Parameter options.

    ToDo: compartmentalize the filtering and/or argument checks.
    """
    from datetime import datetime

    #  Initialize the filter_list
    filter_list = []

    # First, check the arguments
    if time is None and parameter is None:
        insufficient_input_range_select()
        logging.warning("Neither Time nor Parameter provided")
        return kp
    elif time is None:
        # Then only subset based on parameters
        # Need to check whether one or several Parameters given
        inst = []
        obs = []
        # parameter is not None in this case
        if not isinstance(parameter, list):
            if isinstance(parameter,str) or isinstance(parameter,int):
                parameter = [parameter]
            else:
                logging.warning("*****ERROR*****")
                logging.warning("Cannot identify given parameter: %s" % parameter)
                logging.warning("Suggest using param_list(kp) to identify Parameter")
                logging.warning("by index or by name")
                logging.warning("Returning complete original data dictionary")
                return kp
        if minimum is None and maximum is None:
            insufficient_input_range_select()
            logging.warning("No bounds set for parameter(S)")
            logging.warning("Returning complete original data dictionary")
            return kp

        if minimum is None:
            minimum = np.repeat(-np.inf, len(parameter))
        elif not isinstance(minimum, list):
            minimum = [minimum]
        if maximum is None:
            maximum = np.repeat(np.inf, len(parameter))
        elif not isinstance(maximum, list):
            maximum = [maximum]

        if (len(parameter) != len(minimum)) or (len(parameter) != len(maximum)):
            logging.warning("*****ERROR*****")
            logging.warning("---range_select---")
            logging.warning("Number of minima and maxima provided")
            logging.warning("MUST match number of Parameters provided")
            logging.warning("You provided %4d Parameters" % len(parameter))
            logging.warning("             %4d minima" % len(minimum))
            logging.warning("         and %4d maxima" % len(maximum))
            logging.warning("Returning complete original data dictionary")
            return kp

        for param in parameter:
            if not (isinstance(param, int) or isinstance(param, str)):
                logging.warning("*****ERROR*****")
                logging.warning("Cannot identify given parameter: %s" % param)
                logging.warning("Suggest using param_list(kp) to identify Parameter")
                logging.warning("by index or by name")
                logging.warning("Returning complete original data dictionary")
                return kp
            else:
                a, b = get_inst_obs_labels(kp, param)
                inst.append(a)
                obs.append(b)
    # Should I move this below the Time conditional and move
    # Baselining of Filter List to above time

    else:
        # Time has been provided as a filtering agent
        # Determine whether Time is provided as strings or orbits
        if len(time) != 2:
            if parameter is not None:
                logging.warning("*****WARNING*****")
                logging.warning("Time must be provided as a two-element list")
                logging.warning("of either strings (yyyy-mm-dd hh:mm:ss) ")
                logging.warning("or orbits.  Since a Parameter *was* provided,")
                logging.warning("I will filter on that, but ignore the time input.")
                inst = []
                obs = []
                for param in parameter:
                    a, b = get_inst_obs_labels(kp, param)
                    inst.append(a)
                    obs.append(b)

            else:
                # Cannot proceed with filtering
                insufficient_input_range_select()
                logging.warning("Time is not a 2-element list of strings or orbit numbers, no parameter given.")
                logging.warning("Returning complete original data dictionary")
                return kp
        else:
            # We have a two-element Time list: parse it
            if type(time[0]) !=  type(time[1]):
                if parameter is not None:
                    logging.warning("*****WARNING*****")
                    logging.warning("Both elements of time must be same type")
                    logging.warning("Only strings of format yyyy-mm-dd hh:mm:ss")
                    logging.warning("or integers (orbit numbers) are allowed.")
                    logging.warning("Ignoring time inputs; will filter ONLY")
                    logging.warning("on Parameter inputs.")
                else:
                    logging.warning("*****ERROR*****")
                    logging.warning("Both elements of Time must be same type")
                    logging.warning("Only Strings of format yyyy-mm-dd hh:mm:ss")
                    logging.warning("or integers (orbit numbers) are allowed.")
                    logging.warning("Returning original unchanged data dictionary")
                    return kp
            elif type(time[0]) is int:
                # Filter based on orbit number
                min_t = min(time)
                max_t = max(time)
                filter_list.append(kp["Orbit"] >= min_t)
                filter_list.append(kp["Orbit"] <= max_t)
            elif isinstance(time[0], str):
                # Filter acc to string dat, need to parse it
                time_dt = [datetime.strptime(i, "%Y-%m-%d %H:%M:%S") for i in time]
                min_dt = min(time_dt)
                max_dt = max(time_dt)
                kp_dt = [
                    datetime.strptime(i, "%Y-%m-%dT%H:%M:%S") for i in kp["TimeString"]
                ]
                delta_tmin = np.array([(i - min_dt).total_seconds() for i in kp_dt])
                delta_tmax = np.array([(i - max_dt).total_seconds() for i in kp_dt])
                filter_list.append(delta_tmin >= 0)
                filter_list.append(delta_tmax <= 0)
            else:
                # Time provided as other than string or Integer
                if parameter is not None:
                    logging.warning("*****WARNING*****")
                    logging.warning("Times must be strings or integer or orbit numbers")
                    logging.warning("Ignoring time inputs; will filter ONLY")
                    logging.warning("on Parameter inputs.")
                else:
                    logging.warning("*****ERROR*****")
                    logging.warning("Times must be either strings or integer orbit numbers.")
                    logging.warning("Returning original unchanged data dictionary")
                    return kp
            # Now, we apply the Parameter selection
            inst = []
            obs = []
            if isinstance(parameter, int) or isinstance(parameter, str):
                # Then we have a single Parameter to filter on
                # Verify that bounds info exists
                if minimum is None and maximum is None:
                    insufficient_input_range_select()
                    logging.warning("No bounds set for parameter %s" % parameter)
                    logging.warning("Applying only Time filtering")
                    parameter = None
                elif minimum is None:
                    minimum = [-np.inf]  # Unbounded below
                elif maximum is None:
                    maximum = [np.inf]  # Unbounded above
                else:
                    pass  # Range fully bounded
                if not isinstance(minimum, list):
                    minimum = [minimum]
                if not isinstance(maximum, list):
                    maximum = [maximum]
                if len(minimum) != 1 or len(maximum) != 1:
                    logging.warning("*****ERROR*****")
                    logging.warning("---range_select---")
                    logging.warning("Number of minima and maxima provided")
                    logging.warning("MUST match number of Parameters provided")
                    logging.warning("You provided %4d Parameters" % 1)
                    logging.warning("             %4d minima" % len(minimum))
                    logging.warning("         and %4d maxima" % len(maximum))
                    logging.warning("Filtering only on Time")
                    parameter = None
                else:
                    if parameter is not None:
                        a, b = get_inst_obs_labels(kp, parameter)
                        inst.append(a)
                        obs.append(b)
                        nparam = 1  # necessary?
            elif type(parameter) is list:
                if minimum is None and maximum is None:
                    insufficient_input_range_select()
                    logging.warning("No bounds set for parameter list")
                    logging.warning("Applying only Time filtering")
                    parameter = None
                    lmin = 0
                    lmax = 0
                elif minimum is None:
                    minimum = np.repeat(-np.inf,len(parameter))
                    lmin = len(parameter)
                    lmax = len(maximum)
                elif maximum is None:
                    maximum = np.repeat(np.inf, len(parameter))
                    lmin = len(minimum)
                    lmax = len(parameter)
                else:
                    lmin = len(minimum)
                    lmax = len(maximum)
                if (parameter is not None) and (len(parameter) != lmin or len(parameter) != lmax):
                    logging.warning("*****ERROR*****")
                    logging.warning("---range_select---")
                    logging.warning("Number of minima and maxima provided")
                    logging.warning("MUST match number of Parameters provided")
                    logging.warning("You provided %4d Parameters" % len(parameter))
                    logging.warning("             %4d minima" % len(minimum))
                    logging.warning("         and %4d maxima" % len(maximum))
                    logging.warning("Filtering only on Time")
                    parameter = None
                elif parameter is not None:
                    nparam = len(parameter)
                    for param in parameter:
                        a, b = get_inst_obs_labels(kp, param)
                        inst.append(a)
                        obs.append(b)

    # Now, apply the filters
    if parameter is not None:
        inst_obs_minmax = list(zip(inst, obs, minimum, maximum))
        for inst, obs, min_obs, max_obs in inst_obs_minmax:
            filter_list.append(kp[inst][obs] >= min_obs)
            filter_list.append(kp[inst][obs] <= max_obs)

    # Filter list built, apply to data
    filter = np.all(filter_list, axis=0)
    new = {}
    for i in kp:
        temp = kp[i]
        new.update({i: temp[filter]})
    return new


def insufficient_input_range_select():
    """
    This error message is called if user calls range_select with
    inputs that result in neither a valid Time range nor a valid
    Parameter range capable of being determined

    ToDo: Is there a way to hide this from the help feature?
    """
    logging.warning("*****ERROR*****")
    logging.warning("Either a time criterion with two values.")
    logging.warning("  or a parameter name with maximum and/or")
    logging.warning("  minimum values must be provided.")
    logging.warning("Returning the complete original data dictionary")


def get_inst_obs_labels(kp, name):
    """
    Given parameter input in either string or integer format,
    identify the instrument name and observation type for use
    in accessing the relevant part of the data structure
    E.g.: 'LPW.EWAVE_LOW_FREQ' would be returned as
          ['LPW', 'EWAVE_LOW_FREQ']

    Parameters:
        kp: insitu kp data structure/dictionary read from file(s)
        name: string identifying a parameter.
            (Indices must be converted to inst.obs strings before
             calling this routine)
    Output:
        inst (1st arg): instrument identifier
        obs (2nd arg): observation type identifier
    """

    # Need to ensure name is a string at this stage
    name = "%s" % name

    # Now, split at the dot (if it exists)
    tags = name.split(".")
    # And consider the various possibilities...
    if len(tags) == 2:
        return tags
    elif len(tags) == 1:
        try:
            int(tags[0])
            return (find_param_from_index(kp, tags[0])).split(".")
        except:
            logging.warning("*****ERROR*****")
            logging.warning("%s is an invalid parameter" % name)
            logging.warning("If only one value is provided, it must be an integer")
            return [],[]
    else:
        logging.warning("*****ERROR*****")
        logging.warning("%s is not a valid parameter" % name)
        logging.warning("because it has %1d elements" % len(tags))
        logging.warning('Only 1 integer or string of form "a.b" are allowed.')
        logging.warning("Please use .param_list attribute to find valid parameters")
        return [],[]


def find_param_from_index(kp, index):
    """
    Given an integer index, find the name of the parameter

    Parameters:
        kp: insitu kp data structure/dictionary read from file(s)
        index: the index of the desired parameter (integer type)
    Output:
        A string of form <instrument>.<observation>
        (e.g., LPW.EWAVE_LOW_FREQ)
    """

    index = "#%3d" % int(index)
    plist = param_list(kp)
    found = False
    for i in plist:
        if re.search(index, i):
            return i[5:]  # clip the '#123 ' string
    if not found:
        logging.warning("*****ERROR*****")
        logging.warning("%s not a valid index." % index)
        logging.warning("Use param_list to list options")
        return


